(* Mutual exclusion locks (semaphores).
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires schedsig.ml, threadsig.ml *)

(* notice that the scheduler for enqueuing and dequeueing mutexes
   and conditions need not be the same as the one for the thread manager. *)

abstype (''env,'rs) mutex =
   MUTEX of bool ref * (''env thread,'rs) scheduler
and (''env,'rs) condition =
   CONDITION of (''env,'rs) mutex * (''env thread,'rs) scheduler
with
  fun mutex_manager (|[block,yield,reschedule,...]|
		     : (''env thread,'rest,'rs) thread_manager,
		     new_sched : unit -> (''env thread,'rs) scheduler) =
      let fun mutex () : (''env,'rs) mutex =
	      MUTEX (ref false,new_sched ())
	  fun try_acquire (MUTEX (held,_)) : bool =
	      not (!held) andalso (held := true; true)
	  fun acquire (m as MUTEX (_,q)) : unit =
	      if try_acquire m
		 then ()
	      else block q
	  fun release (MUTEX (held,|[dequeue,enqueue,...]|))
		: unit =
	      case dequeue () of
		   SOME thread => reschedule thread
		 | NONE => held := false
	  fun with_mutex (m : (''env,'rs) mutex,body : (unit -> 'a)) : 'a =
	      (acquire m;
	       body () handle exn => (release m; raise exn))
	      before release m

	  fun condition (m : (''env,'rs) mutex) : (''env,'rs) condition =
	      CONDITION (m,new_sched ())
	  fun with_condition (CONDITION (mutex,_) : (''env,'rs) condition,
			      body : (unit -> 'a)) : 'a =
	      with_mutex (mutex,body)
	  fun awaken (CONDITION (m as MUTEX (_,|[enqueue = putm,...]|),
				 |[dequeue = getc,...]|))
			: bool =
	      case getc () of
		   SOME thread => ((if try_acquire m
				       then reschedule thread
				    else putm thread);
				   true)
		 | NONE => false
	  fun signal cond = (awaken cond; ())
	  fun broadcast cond =
	      while awaken cond do ()
	  fun wait (CONDITION (m,q)) =
	      (release m; block q)
	  fun await (cond,test) =
	      while not (test ())
		do wait cond
      in
	 |[mutex = mutex,
	   try_acquire = try_acquire,
	   acquire = acquire,
	   release = release,
	   with_mutex = with_mutex,
	   condition = condition,
	   mutex_of = fn (CONDITION (m,_)) => m,
	   with_condition = with_condition,
	   signal = signal,
	   broadcast = broadcast,
	   wait = wait,
	   await = await
	   ]|
      end
end
