(* Simulated threads, in the style of Cooper and Morrissett's report
   on adding threads to SML (appendix A).
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires schedsig.ml, threadsig.ml *)

(* provides a type ''env thread of threads with attached environment
   (or environment ID) of type ''env, an exception Deadlock,
   and a function thread_manager that creates a thread manager
   parameterized by a scheduler.
     A thread manager has the following functions:
   - fork : (unit -> unit) -> unit forks a new thread to compute
     the function in argument, which it does immediately after
     rescheduling the current thread.
   - abort : unit -> 'a terminates the current thread.
   - yield : unit -> unit switches to the next thread.

	Threads are non-preemptive, here.
	Moreover, they are simulated by catch/throw, which is
a bit faster than callcc/throw. Actually, continuations are encoded
as threads with shared stack caches in HimML, so this should provide
a relatively fast implementation.
*)

abstype ''env thread = THREAD of unit cont * ''env
	(* a thread is a stack chunk (the unit cont), plus
	   an environment, which may be anything you wish,
	   as long as we can compare them (typically, these
	   are environment IDs). *)
with
  exception Deadlock (* not a real deadlock;
			signals when no thread remains to be scheduled. *)

  fun thread_manager
    (current_env : ''env ref,
     new_env : unit -> ''env)
    (sched as |[enqueue,dequeue,...]|
     : (''env thread,'rest) scheduler) =
    let fun thread k = THREAD (k,!current_env)
	fun run_next () =
	    case dequeue () of
		SOME (THREAD (k,env)) => (current_env := env;
					  throw k ())
	      | NONE => raise Deadlock
	fun block (|[enqueue,...]| : (''env thread,'rest) scheduler) =
	(* switch to next thread, after remembering the current one
	   in the designated queue;
	   we can use catch instead of callcc since we
	   know that threads can only be relaunched once. *)
	    catch (fn k => (enqueue (thread k);
			    run_next ()))
	fun fork f = (* fork a new thread; we use catch instead of callcc again. *)
	    catch (fn k => (enqueue (thread k);
			    current_env := new_env ();
			    f () handle exn => (#put stderr "Unhandled exception ";
						print stderr (pack exn);
						#put stderr " raised in thread.\n";
						#flush stderr ());
			    run_next ()))
    in
	|[abort = run_next, (* named exit by Cooper and Morrissett *)
	  block = block,
	  yield = fn () => block sched,
	  fork = fork,
	  reschedule = enqueue,
	  deschedule = dequeue
	  ]|
    end
end
