(* Applicative Union-Find structures.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

type (''a,'b,'rest) union_find_tree =
    |[link : ''a -m> ''a, (* links going up towards roots (aka colors) *)
      info : ''a -m> (* maps from colors to data *)
      |[card : num, (* the number of elements of this color *)
	... : 'rest]|, (* and other data *)
       userdata : 'b]| (* and other fields ... *)

fun empty_union_find_tree userdata =
    |[link = {}, info = {}, userdata = userdata]|

(* find : (''a,'b,'rest) union_find_tree -> ''a -> ''a *)
(* given a tree, find computes the color of an object *)
fun find |[link,...]| =
  let fun f x = if x inset link
		    then f (?link(x))
		(* recursivite terminale, garantie n'utilisant pas de pile *)
		else x
  in
      f (* currifie, comme ca la recursion est plus rapide *)
  end

(* merge : ('b -> 'b) * (|[card : num,... : 'rest]| *
 |[card : num,... : 'rest]| -> |[card : num,... : 'rest]|) ->
 (''a,'b,'rest) union_find_tree -> ''a * ''a -> (''a,'b,'rest) union_find_tree *)
(* given a function extra, which may do additional merging operations
 on the 'rest fields, a function cextra, which does the same on the
 'crest fields and a tree, merge colors x1 and x2,
 yielding a new tree *)
fun merge (extra,cextra) (tree as |[link,info,userdata]|) (x1,x2) =
    if x1=x2 then tree else
	let val i1 as |[card = c1,...]| = ?info x1
	    and i2 as |[card = c2,...]| = ?info x2
	    val tree' = if c1<c2
			    then |[link = link ++ {x1 => x2},
				   info = ({x1} <-| info) ++
				   {x2 => cextra (i1,i2)},
				   userdata = extra userdata]|
			else |[link = link ++ {x2 => x1},
			       info = ({x2} <-| info) ++
			       {x1 => cextra (i2,i1)},
			       userdata = extra userdata]|
	in
	    tree'
	end

(* color_card : (''a,'b,'rest) union_find_tree -> ''a -> num *)
(* given a tree, computes the number of elements in the class of the color x *)
fun color_card |[info,...]| x = #card (?info(x))+1

    (* a typical additional field in each color is a set of terms having this
     color; instead of coding them with sets, we could use lists, but for a
     better efficiency, we choose heaps: *)
datatype ''a heap =
    H0
  | H1 of ''a
  | H2 of ''a heap * ''a heap

fun concat_heap (H0,h) = h
  | concat_heap (h,H0) = h
  | concat_heap hpair = H2 hpair

fun map_heap f =
    let fun map H0 = H0
	  | map (H1 x) = H1 (f x)
	  | map (H2 (h,h')) = H2 (map h,map h')
    in
	map
    end

fun app_heap f =
    let fun app H0 = ()
	  | app (H1 x) = f x
	  | app (H2 (h,h')) = (app h;
			       app h')
    in
	app
    end
