/* Definitions for the bytecode interpreter.
   Copyright (C) 1995-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2021 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

typedef struct byteCodeChunk {
  CHAR kind; /* = E_BYTECODE */
  CHAR mark;
  struct byteCodeChunk *next;
  COUNTER nargs; /* initial number of arguments (not counting environment)
		    in initial stack. */
  COUNTER ntargs; /* arity, as given by typing, -1 if don't know. */
  COUNTER stackSize; /* max number of elements on run-time stack */
  INT mapStackOff; /* number of bytes from start of run-time gcStackView to end of
		      run-time stack (=start of stack of pointers private to map iterations). */
  INT excStackOff; /* number of bytes from start of run-time gcStackView to end of map stack
		      (=start of stack of excHandler's) */
  COUNTER nConsts; /* number of constant ExprPtrs that we need */
  CHAR *rawData; /* pointer to beginning of rawdata section */
  OPCODE *instr; /* pointer to beginning of instructions */
  OPCODE *debug; /* pointer to beginning of debug instructions */
  COUNTER size; /* number of bytes in structure */
  struct Expression *consts[1]; /* we have nConsts ExprPtrs here */
} byteCodeChunk;

#define BC_STACK_SIZE 512

/* Utilities for outputting bytecode: */
/* opcodes: */
#define BC_BUFSIZE 256

typedef struct bcBuffer {
  INT n;
  struct bcBuffer *next;
  OPCODE instr[BC_BUFSIZE];
} bcBuffer;

/* boxed constant data: */
#define BCC_BUFSIZE 128
typedef struct bccBuffer {
  INT n;
  struct bccBuffer *next;
  ExprPtr data[BCC_BUFSIZE];
} bccBuffer;

/* unboxed data: */
#define BCU_BUFSIZE 1024
typedef struct bcuBuffer {
  INT n;
  struct bcuBuffer *next;
  CHAR data[BCU_BUFSIZE];
} bcuBuffer;

typedef struct bcFile {
  INT nargs;
  INT stackDepth,maxStackDepth;
  INT mapStackDepth,maxMapStackDepth;
  COUNTER ntargs; /* arity, as given by typing, -1 if don't know. */
  COUNTER dead; /* true if all code that would be output would be dead */
  bcBuffer *buf,*current; /* buffer of OPCODEs */
  bcBuffer *dbuf,*dcurrent; /* buffer for debugging codes */
  bccBuffer *cbuf,*ccurrent; /* buffer for ExprPtr constants */
  ExprPtr cbuf_e; /* handle to cbuf, for gc purposes. */
  INT cpos;
  bcuBuffer *ubuf,*ucurrent; /* buffer for raw data */
#if DEBUGGING!=0
  COUNTER lastLine, lastPos;
  ExprPtr lastDebugZone;
#endif
} bcFile;

typedef struct compEnv {
  COUNTER level; /* level of the current environment (0=toplevel, ie
		    only the global environment is present) */
  COUNTER ilevel; /* same, but for varintro's */
  INT globenvs; /* bit-mask for those levels which are known to be mere copies
		   of GLOBENV (useful with modules, which do build such fake copies) */
  INT flags;
#define CEF_NEEDENV 1
#define CEF_SPARESTACK 2
  struct Var me; /* used for detecting tail-recursions */
  /* true if environments must really be compiled */
  struct _compEnv { /* this structure should look like a gcStackView */
    COUNTER n;
    struct gcStackView *next;
#define CE_N 9
    ExprPtr p[CE_N];
  } _comp;
#define CE_VARS(ce) (ce)->_comp.p[0]
  /* map from couples (i,level), or i if level=0, to positions in
     the evaluation stack where we can find the same variable;
     level is from the top, and positions in the stack are from
     the start (not the top) of the stack.
     As a particular case, couples (NIL,level) denote the environment
     at level level itself. */
#define CE_INLINES(ce) (ce)->_comp.p[1]
  /* map from couples as above to E_BYTECODE expressions to inline */
#define CE_HANDLERS(ce) (ce)->_comp.p[2]
  /* stack of exception handlers, as a list.
     The entries are E_COUNTED (E_COUNTED (merge-node, stack-level),
     mapstack-level), where
     merge nodes are described under gotop and errp below. */
#define CE_LETDELAY(ce) (ce)->_comp.p[3]
  /* only valid when compiling patterns, holds a map
     from ranks to stack indexes, that should then be
     translated to BC_LET assignment expressions
     after the pattern is compiled. */
#define CE_DZ(ce) (ce)->_comp.p[4]
  /* current debugZone or NIL */
#define CE_PROF(ce) (ce)->_comp.p[5]
  /* EA_PROFILE build header structure for current function, or NIL. */
#define CE_FV(ce) (ce)->_comp.p[6]
  /* map from E_POLYFUN nodes to sets of free variables (as in CE_VARS
     and CE_INLINES) */
#define CE_DBGNAMES(ce) (ce)->_comp.p[7]
  /* map from ranks to a debugging info E_LET(name,body), or
     E_GOTO(name,body), or E_APPL(name,body) (see debug.h). */
#define CE_PARENT_DBGNAMES(ce) (ce)->_comp.p[8]
  /* similar, cumulating variable info coming from the closure environment
   of the current code.  This maps names to E_LET(rank,body),
   or E_GOTO(rank,body), or E_APPL(rank,body). */
} compEnv;

/* Compiling merge nodes is done as follows.
   As far as bytecodes are concerned, merge nodes will be represented by
   a do-nothing BC_MERGE opcode.
   When compiling, however, we use (in variables named like gotop or errp) a cons
   whose car points to:
   1) either (forward jump) a list of varrefs (rank = address of goto,
      level = stackDepth when goto was triggered)
   2) or (backward jump) a varref (rank = address of merge node,
      level = stackDepth at merge node)
   The cdr is usually NIL, but can also point to other gotop nodes (in compiling
   patterns with choices, notably).
*/

/* data types for compiling iterators and quantifers: */

typedef void (*compileNeutralFunc) P((IN bcFile *f,IN CONST compEnv *env));
     /* --- neutral (if !quantifier); --- (if quantifier) */
typedef void (*compileCombineFunc) P((IN bcFile *f,IN CONST compEnv *env,
				      IN COUNTER l1));
     /* an ... a2 a1 --- an ... a2 a'1 ; where l1 is the stack level of an */
typedef void (*compileCombine1Func) P((IN bcFile *f,IN CONST compEnv *env));
     /* a --- a' ; or goto errp after pushing the result on stack of depth errlevel. */
typedef void (*compileEvalPromoteFunc) P((IN bcFile *f,IN CONST compEnv *env,
					  IN CONST ExprPtr body,
					  IN COUNTER errlevel,IN COUNTER maperrlevel,
					  INOUT ExprPtr errp));
     /* --- a (if !quantifer); --- (if quantifier) */
typedef void (*compileFinalizeFunc) P((IN bcFile *f,IN CONST compEnv *env));
     /* a --- a' ; function called when iteration has been executed at least once,
      and has terminated normally. */

struct compilesimpleof {
  compileNeutralFunc neutral;
  compileCombineFunc combine;
  compileCombine1Func combine1;
  compileEvalPromoteFunc evalpromote;
  compileFinalizeFunc finalize;
  ExprPtr dompats;
  ExprPtr domains;
  ExprPtr locals;
  ExprPtr pat;
  ExprPtr body,pred;
  COUNTER nlocals;
  Boolean mapsweep;
  COUNTER resclass; /* we say that quantifier==(resclass!=0)
		       This is 0 if program is an expression, 1 if it is an
		       existential test, 2 if it is a universal test,
		       3 if it is a statement (affects how stack is used to
		       return result) */
  Boolean simplep; /* true to compile loops as just loops, false
		      to compile them as a first loop dealing with
		      the first element, then as a second loop dealing
		      with the rest. */
};

typedef void (*compileBodyFunc) P((IN bcFile *f,IN CONST compEnv *env,
				   IN CHAR *data,INOUT ExprPtr failp,
				   IN COUNTER errlevel,IN COUNTER errmaplevel,
				   INOUT ExprPtr errp));
     /* x ---  <...> res (if !quantifier); x --- <...> (if quantifier) */

typedef void (*compileFunc) P((IN bcFile *f,IN CONST compEnv *env,
			       IN CONST ExprPtr e,INOUT ExprPtr wrongp));

typedef struct applyData {
  bcFile *f;
  compEnv *env;
  ExprPtr fun;
  compEnv *arg_env;
  ExprPtr arg,otherargs;
  Boolean tail;
  compileFunc compile;
  Boolean push;
  ExprPtr wrongp;
} applyData;

#define APPLYDATA (applyData *)arg
#define AD_F ((applyData *)arg)->f
#define AD_ENV ((applyData *)arg)->env
#define AD_FUN ((applyData *)arg)->fun
#define AD_ARG_ENV ((applyData *)arg)->arg_env
#define AD_ARG ((applyData *)arg)->arg
#define AD_OTHERS ((applyData *)arg)->otherargs
#define AD_TAIL ((applyData *)arg)->tail
#define AD_COMPILE ((applyData *)arg)->compile
#define AD_PUSH ((applyData *)arg)->push
#define AD_WRONGP ((applyData *)arg)->wrongp

/* bytecodes: */

#define BC_ILLEGAL 0
#define BC_CGLOBAL 1
/* CGLOBAL <index-of-global> : --- value-of-global */
#define BC_CONST 2
/* CONST <i> : --- consts[i] */
#define BC_CLOSURE 3
/* CLOSURE <i> : env --- closure(consts[i],env) */
#define BC_REF 4
/* REF : v --- ref(v) */
#define BC_KTUPLE1 5
/* KTUPLE1 : v --- Ktuple1(v) */
#define BC_MAKETUPLE1 6
/* MAKETUPLE_1 : v --- MakeTuple1(v) */
#define BC_KONS 7
/* KONS : a b --- Kons(a,b) */
#define BC_CONS 8
/* CONS : a b --- Cons(a,b) */
#define BC_FETCH 9
/* FETCH <i> : an ... a2 a1 --- an ... a2 a1 a[-i] */
#define BC_KRECORD1 10
/* KRECORD1 <i> : a --- Krecord1(consts[i] as shape, a) */
#define BC_MAKERECORD1 11
/* KRECORD1 <i> : a --- MakeRecord1(consts[i] as shape, a) */
#define BC_KTUPLE2 12
/* KTUPLE2 : a b --- Ktuple2(a,b) */
#define BC_MAKETUPLE2 13
/* MAKETUPLE2 : a b --- MakeTuple2(a,b) */
#define BC_KTUPLEGNRL 14
/* KTUPLEGNRL <n> : an ... a2 a1 --- Ktuple(an, ..., a2, a1) */
#define BC_MAKETUPLEGNRL 15
/* MAKETUPLEGNRL <n> : an ... a2 a1 --- MakeTuple(an, ..., a2, a1) */
#define BC_KRECORDGNRL 16
/* KRECORD_GNRL <n> <i> : an ... a2 a1 --- Krecord(shape = consts[i], an, ..., a2, a1) */
#define BC_MAKERECORDGNRL 17
/* MAKERECORD_GNRL <n> <i> : an ... a2 a1 --- MakeRecord(shape = consts[i], an, ..., a2, a1) */
#define BC_NOEXPR 18
/* NOEXPR <n> : --- NOEXPR ... NOEXPR (n times) */
#define BC_TUCK 19
/* TUCK <i> : an a[n-1] ... a2 a1 x --- x a[n-1] ... a2 a1, where i=-n (and an==NOEXPR) */
#define BC_MAPLET 20
/* MAPLET : a b --- Maplet(a => b) */
#define BC_OVERWRITE 21
/* OVERWRITE : a b --- a++b */
#define BC_UNDERWRITE 22
/* UNDERWRITE : a b --- b++a */
#define BC_VARREF0 23
#define BC_VARREF1 24
#define BC_VARREF2 25
#define BC_VARREF3 26
/* VARREF[l] <i> <j> : an ... a2 a1 --- an ... a2 a1 varref(level=l [from env=a[-i]], rank=j) */
#define BC_VARREF 27
/* VARREF <i> <l> <j> : an ... a2 a1 --- an ... a2 a1 varref(level=l [from env=a[-i]], rank=j) */
#define BC_GLOBVARREF 28
/* GLOBVARREF <j> : --- varref(rank = j in GLOBENV) */
#define BC_DROP 29
/* DROP <n> : an ... a2 a1 --- */
#define BC_MERGE 30
/* MERGE : --- ; do nothing (denotes forward merge nodes in the CFG) */
#define BC_GOTO 31
/* GOTO <n> <i> : --- move program counter by <i> opcodes (relative branch), drop stack by <n> */
#define BC_RAISE 32
/* RAISE : exn --- value ; raise exception, debugger may return value */
#define BC_FAKEENV 33
/* FAKEENV : env --- env' ; cons fake copy of GLOBENV in front of environment env */
#define BC_CONSENV 34
/* CONSENV <n> : env --- env' ; cons local environment with n free slots to env */
#define BC_LET 35
/* LET <i> <k1> <j1> ... <kn> <jn> 0 : an ... a2 a1 --- an ... a2 a1
   ; set slot j of env=a[-i] to a[-k], for every couple j,k in the list */
#define BC_SWDROP 36
/* SWDROP <i> : an ... a2 a1 --- a1 ; where n=-i */
#define BC_QENV 37
/* QENV : a --- QENV(a) */
#define BC_CAR 38
/* CAR : (a,b) --- a */
#define BC_CDR 39
/* CDR : (a,b) --- b */
#define BC_IS 40
/* IS <n> <i> <j> : a --- a ; goto <i> (and drop <n>) if a is not consts[j], otherwise continue */
#define BC_REFVAL 41
/* REFVAL : ref(a) --- a */
#define BC_ISEQUAL 42
/* ISEQUAL <n> <i> : a b --- ; goto <i> (+drop <n>) if a!=b */
#define BC_NTH 43
/* NTH <i> : (a0,...,an) --- ai */
#define BC_TYPEMATCH 44
/* TYPEMATCH <n> <i> <j> : type --- ; goto <i> (+drop <n>) if TypeMatch(type,consts[j]) is true (no match) */
#define BC_CARDEQ 45
/* CARDEQ <n> <i> <k> : a --- a ; goto <i> (+drop <n>) if card a<>k */
#define BC_CARDGEQ 46
/* CARDGEQ <n> <i> <k> : a --- a ; goto <i> (+drop <n>) if card a<k */
#define BC_SELECT 47
/* SELECT <i> : record(a0,...,an) --- aj ; such that j is the index for label consts[i] */
#define BC_ISNEMAP 48
/* ISNEMAP <n> <i> : a --- a ; goto <i> (+drop <n>) if a={} */
#define BC_MAPGET 49
/* MAPGET <n> <i> : m x --- y ; find y = ?m x, or goto <i> (+drop <n>) if x not in m */
#define BC_MAPREMOVE 50
/* MAPREMOVE : x m --- m \ {x} */
#define BC_RETURN 51
/* RETURN <n> : a --- ? ; return value a from bcEval() (and fake a SWDROP <-n-1> before,
   where <n+1> is the height of the stack) */
#define BC_MAPADD 52
/* MAPADD : x y m --- m ++ {x => y} */
#define BC_SETCAR 53
/* SETCAR : a b --- ; sets car(b) to a */
#define BC_CALL 54
/* CALL : f a --- f(a) */
#define BC_N_CALL 55
/* N_CALL <n> : f a1 ... an --- f(a1,...,an) */
/* Was formerly attributed to:
   TAILCALL <i> : f a --- ? ; return f(a) (and fake a SWDROP <-1-i>, i.e. the height
   of the stack is <i+2>) */
#define BC_SETREF 56
/* SETREF : r a --- ; set ref r to value a */
#define BC_LOOP 57
/* LOOP : --- ; do nothing (except checking for ctrl-C, and so on;
   denotes backward merge nodes in the CFG) */
#define BC_KMU1 58
/* KMU1 <i> : rec v --- rec ++|[lab = v]| ; where lab = consts[i] */
#define BC_MU1 59
/* MU1 <i> : rec v --- rec ++|[lab = v]| ; where lab = consts[i], result is not shared */
#define BC_KMUGNRL 60
/* KMUGNRL : rec mod --- rec++mod */
#define BC_MUGNRL 61
/* MUGNRL : rec mod --- rec++mod ; result record is not shared */
#define BC_IF 62
/* IF <n> <i> : b --- ; goto <i> (+drop <n>) if b is false */
#define BC_SWAP 63
/* SWAP : a b --- b a */
#define BC_NEWQUEUE 64
/* NEWQUEUE : --- empty-queue */
#define BC_CONCATQUEUE 65
/* CONCATQUEUE <i> : an ... a2 a1 --- an ... a2 a'1 ; a'1 = a[-i] @ a1,
   destroy both a1 and a[-i] */
#define BC_NEW1QUEUE 66
/* NEW1QUEUE : a --- queue(a) */
#define BC_QUEUELIST 67
/* QUEUELIST : queue --- list */
#define BC_MAPINTER 68
/* MAPINTER : s1 s2 --- s1 & s2 */
#define BC_MAPDOMRESTRTO 69
/* MAPDOMRESTRTO : s m --- s <| m */
#define BC_STARTLISTSWEEP 70
/* STARTLISTSWEEP <i> : an ... a2 a1 --- an ... a2 a1 a[-i] ; just like FETCH. */
#define BC_STARTMAPSWEEP 71
/* STARTMAPSWEEP <i> : an ... a2 a1 --- an ... a2 a1 first-elt NOEXPR ... NOEXPR
   (MAP_CHAIN_LENGTH of them); also push the address of the first NOEXPR on the map stack. */
#define BC_STARTSUBMAPSWEEP 72
/* STARTSUBMAPSWEEP <i> : an ... a2 a1 --- an ... a2 a1 {} InitialSubMap(a[-i]) */
#define BC_GETLISTSWEEP 73
/* GETLISTSWEEP <i> : an ... a2 a1 --- an ... a2 a1 car(a[-i]) ; just like FETCH;CAR */
#define BC_GETMAPSWEEP 74
/* GETMAPSWEEP <i> : an ... a2 a1 --- an ... a2 a1 a[-i] ; just like FETCH. */
#define BC_GETSUBMAPSWEEP 75
/* GETSUBMAPSWEEP <i> : an ... a2 a1 --- an ... a2 a1 a[-i] ; just like FETCH. */
#define BC_INCLISTSWEEP 76
/* INCLISTSWEEP <n> <i> <k> : ak ... a1 --- ak ... a1 ; increment list ak,
   and goto <i> (+drop <n>) if end of list. */
#define BC_INCMAPSWEEP 77
/* INCMAPSWEEP <n> <i> <k> <j> : ak ... a1 --- ak ... a1 ; replace ak by next map pointed to
   by maptop[-j] and maptop[-j+1], and goto <i> (+drop <n>) if end of map. */
#define BC_INCSUBMAPSWEEP 78
/* INCSUBMAPSWEEP <n> <i> <k> : ak ... a1 --- ak ... a1 ; replace ak by next map pointed to
   by maptop[j] and maptop[j+1], and goto <i> (+drop <n>) if end of map. */
#define BC_LISTQUEUE 79
/* LISTQUEUE : list --- queue ; convert a list to a queue ; (converse of QUEUELIST) */
#define BC_QUEUEUNSHAREDLIST 80
/* QUEUEUNSHAREDLIST : queue --- list ; where list is not shared */
#define BC_HANDLE 81
/* HANDLE <n> <i> : --- ; push an exception frame;
   on raise: --- exn ; and goto <i> (+drop <n>) */
#define BC_NEXTHANDLE 82
/* NEXTHANDLE : --- ; pop current exception frame */
#define BC_EXCEPTION 83
/* EXCEPTION : a --- exc(a) ; build new exception with internal value a */
#define BC_DELAY 84
/* DELAY : a --- promise(a) */
#define BC_PACK 85
/* PACK : val type --- pack(val,type) */
#define BC_GOTO_0 86
/* GOTO_0 <n> <i> ; same as BC_GOTO (n=0) */
#define BC_MAPDROP 87
/* MAPDROP <n> : --- ; drop <n> from mapstack */
#define BC_FUNCTION 88
/* FUNCTION : --- ; do nothing (except checking for ctrl-C, and so on;
   denotes function entry points) */
#define BC_IADD 89
/* IADD : a b --- a+b ; add two integers */
#define BC_ISUB 90
/* ISUB : a b --- a-b ; subtract two integers */
#define BC_IMUL 91
/* IMUL : a b --- a*b ; multiply two integers */
#define BC_IDIV 92
/* IDIV : a b --- a div b ; get quotient (it is always the case that b<>0) */
#define BC_IMOD 93
/* IMOD : a b --- a mod b ; get remainder (it is always the case that b<>0) */
#define BC_ISNZERO 94
/* ISNZERO <n> <i> : a --- a ; goto <i> (+drop <n>) if a is zero */
#define BC_IOR 95
/* IOR : a b --- a|b */
#define BC_IXOR 96
/* IXOR : a b --- a^b */
#define BC_IAND 97
/* IAND : a b --- a&b */
#define BC_ILSL 98
/* ILSL : a b --- a<<b */
#define BC_ILSR 99
/* ILSR : a b --- a>>b ; where a is treated as unsigned */
#define BC_IASR 100
/* IASR : a b --- a>>b ; where the sign of a is kept and extended */
#define BC_IABS 101
/* IABS : a --- |a| */
#define BC_INEG 102
/* INEG : a --- -a */
#define BC_ISQR 103
/* ISQR : a --- a*a */
#define BC_INOT 104
/* INOT : a --- ~a */
#define BC_ILT 105
/* ILT <n> <i> : a b --- ; goto <i> (+drop <n>) if !a<b */
#define BC_ILE 106
/* ILE <n> <i> : a b --- ; goto <i> (+drop <n>) if !a<=b */
#define BC_IGT 107
/* IGT <n> <i> : a b --- ; goto <i> (+drop <n>) if !a>b */
#define BC_IGE 108
/* IGE <n> <i> : a b --- ; goto <i> (+drop <n>) if !a>=b */
#define BC_IMIN 109
/* IMIN : a b --- min(a,b) */
#define BC_IMAX 110
/* IMAX : a b --- max(a,b) */
#define BC_NUMINT 111
/* NUMINT : a --- num(a) ; convert integer to floating-point */
#define BC_INTNUM 112
/* INTNUM : a --- int(a) ; convert floating-point to integer */
#define BC_SIZE 113
/* SIZE : s --- len(s) ; compute length of string */
#define BC_CALLBIND 114
/* CALLBIND <i> : env arg --- res ; call bytecode const[i] so as to update environment env
   (used to import modules); res is a Boolean, which is true if some binding failed. */
#define BC_LET1 115
/* LET1 <i> <k> <j> : an ... a2 a1 --- an ... a2 a1
   ; set slot j of env=a[-i] to a[-k] (specialized version of BC_LET for n=1) */
#define BC_CLOSURE0 116
/* CLOSURE0 <i> : --- closure(consts[i],NIL) (for closures without free variables) */
#define BC_AREF 117
/* AREF : a b --- a[b] */
#define BC_ASET 118
/* ASET : a b c --- ; store c in a[b]. */
#define BC_DATA 119
/* DATA <m> <i> : a --- E_DATA(ttm=consts[i],a,magic=m), unshared. */
#define BC_KDATA 120
/* KDATA <m> <i> : a --- E_DATA(ttm=consts[i],a,magic=m), shared. */
#define BC_ISDATA 121
/* ISDATA <n> <i> <m> : a --- b ; goto <i> (and drop <n>) if a has magic!=m. */
#define BC_SWITCH 122
/* SWITCH <m+2> <n> <i[-m]> ... <i[n]> : a --- a ; goto <i[j]> if a has magic==j,
   where magic is between -m and n.  We store m+2 instead of m to get slightly more
   speed in the interpreter. */
#define BC_FETCHCAR 123
/* FETCHCAR <i> : an ... a2 a1 --- an ... a2 a1 car(a[-i]) ; just like FETCH;CAR */
#define BC_FETCHCDR 124
/* FETCHCDR <i> : an ... a2 a1 --- an ... a2 a1 cdr(a[-i]) ; just like FETCH;CDR */
#define BC_FETCHREFVAL 125
/* FETCHREFVAL <i> : an ... a2 a1 --- an ... a2 a1 refval(a[-i]) ; just like FETCH;REFVAL */
#define BC_NTH_0 126
#define BC_NTH_1 127
#define BC_NTH_2 128
#define BC_NTH_3 129
#define BC_NTH_4 130
#define BC_NTH_5 131
#define BC_NTH_6 132
#define BC_NTH_7 133
/* NTH_<i> is the same as NTH <i>, 0<=i<8 */
#define BC_CARCDR 134
/* CARCDR : a --- car(a) cdr(a) ; just like FETCH -1;CAR;SWAP;CDR */
#define BC_CDRCAR 135
/* CDRCAR : a --- cdr(a) car(a) ; just like FETCH -1;CDR;SWAP;CAR */
#define BC_3ARGS 136
/* 3ARGS : a --- a(0) a(1) a(2) ; where a is a tuple */
#define BC_FETCH_m8 137
#define BC_FETCH_m7 138
#define BC_FETCH_m6 139
#define BC_FETCH_m5 140
#define BC_FETCH_m4 141
#define BC_FETCH_m3 142
#define BC_FETCH_m2 143
#define BC_FETCH_m1 144
/* FETCH_m<i> is the same as FETCH <-i> */
#define BC_FETCHCAR_m8 145
#define BC_FETCHCAR_m7 146
#define BC_FETCHCAR_m6 147
#define BC_FETCHCAR_m5 148
#define BC_FETCHCAR_m4 149
#define BC_FETCHCAR_m3 150
#define BC_FETCHCAR_m2 151
#define BC_FETCHCAR_m1 152
/* FETCHCAR_m<i> is the same as FETCHCAR <-i> */
#define BC_FETCHCDR_m8 153
#define BC_FETCHCDR_m7 154
#define BC_FETCHCDR_m6 155
#define BC_FETCHCDR_m5 156
#define BC_FETCHCDR_m4 157
#define BC_FETCHCDR_m3 158
#define BC_FETCHCDR_m2 159
#define BC_FETCHCDR_m1 160
/* FETCHCDR_m<i> is the same as FETCHCDR <-i> */
#define BC_FETCHREFVAL_m8 161
#define BC_FETCHREFVAL_m7 162
#define BC_FETCHREFVAL_m6 163
#define BC_FETCHREFVAL_m5 164
#define BC_FETCHREFVAL_m4 165
#define BC_FETCHREFVAL_m3 166
#define BC_FETCHREFVAL_m2 167
#define BC_FETCHREFVAL_m1 168
/* FETCHREFVAL_m<i> is the same as FETCHREFVAL <-i> */
#define BC_BREAK 169
/* BREAK : --- ; enters the debugger (used for breakpoints). */
#define BC_EXCEPTION_BENIGN 170
/* EXCEPTION_BENIGN : a --- exc(a) ; build new exception with internal value a, but
 a benign one (one that the debugger won't try to catch raises of). */
#define BC_MAPDOMRESTRBY 171
/* MAPDOMRESTRBY : s m --- s <-| m */
#define BC_DIFF 172
/* DIFF : s1 s2 --- s1\s2 */
#define BC_DELTA 173
/* DELTA : m1 m2 --- m1 delta m2 */
#define BC_NEXTHANDLE_NODEBUG 174
/* NEXTHANDLE_NODEBUG : --- ; as NEXTHANDLE, but does not test for debugMode. */
#define BC_IAREF 175
/* IAREF : a b --- a[b] */
#define BC_IASET 176
/* IASET : a b c --- ; stores c into a[b]. */
#define BC_RLT 177
/* RLT <n> <i> : a b --- ; goto <i> (+drop <n>) if !a#<b */
#define BC_RLE 178
/* RLE <n> <i> : a b --- ; goto <i> (+drop <n>) if !a#<=b */
#define BC_RGT 179
/* RGT <n> <i> : a b --- ; goto <i> (+drop <n>) if !a#>b */
#define BC_RGE 180
/* RGE <n> <i> : a b --- ; goto <i> (+drop <n>) if !a#>=b */
#define BC_RMIN 181
/* RMIN : a b --- min(a,b) */
#define BC_RMAX 182
/* RMAX : a b --- max(a,b) */
#define BC_FLOOR 183
/* FLOOR : a --- floor(a) */
#define BC_CEIL 184
/* CEIL : a --- ceil(a) */
#define BC_CONSMEMO 185
/* CONSMEMO : a b --- Cons(a,b); just like CONS, except used to compile
   memo functions: the Cons(a,b) is mutable, and its car will be modified
   by SETCAR. */
#define BC_MAPADD2 186
/* MAPADD2 : m x y --- m ++ {x => y} ; same semantics as MAPADD. */
#define BC_ORD 187
/* ORD : s --- s[0] ; may raise StringNth. */
#define BC_ORDOF 188
/* ORDOF : s n --- s[n] ; may raise StringNth. */
#define BC_SUBSTR 189
/* SUBSTR : s i j --- s[i..j-1] ; may raise StringNth. */
#define BC_NCARGET 190
/* NCARGET <n> <i> : (m, _) x --- y ; find y = ?m x, or goto <i> (+drop <n>) if x not in m;
 m is built using NPut() instead of MapPut(). */
#define BC_NCARPUT 191
/* NCARPUT : (m, _) x y --- y; add x=>y to map m, using NPut() */
#define BC_NARGS 192
/* NARGS <n> : a --- a(0) a(1) ... a(n-1); where a is a tuple_gnrl, and n>=3;
   in case n==3, preferrably use BC_3ARGS. */
#define BC_NSWDROP 193
/* NSWDROP <i> <j> : an ... ak+1 ak ... a1 --- ak ... a1 ; where n=-i, k=-j.
 Not recommended if k==0 (use DROP <n>) or if k==1 (use SWDROP <i>). */

#define BC_HASTAG 194
/* HASTAG+k : <n> <i> : a --- a ; goto <i> (+drop <n>) if the tag of a is not k */
#define BC_ISNIL (BC_HASTAG+E_NIL)
#define BC_ISCONS (BC_HASTAG+E_CONS)
#define BC_ISTUPLE2 (BC_HASTAG+E_TUPLE_2)
#define BC_ISRECORD1 (BC_HASTAG+E_RECORD_1)
#define BC_ISEMPTY (BC_HASTAG+E_EMPTY)
#define BC_ISMAPLET (BC_HASTAG+E_MAPLET)
#define BC_ISMAP (BC_HASTAG+E_MAP)

/*** Exported functions: */

IMPORT byteCodeChunk *bcAlloc P((IN INT size));
IMPORT INT getConstant P((IN bcFile *f, IN CONST ExprPtr e));

IMPORT ExprPtr Compile P((IN CONST ExprPtr e,IN COUNTER level,IN Boolean bindp));
IMPORT Boolean hasChoice P((IN CONST ExprPtr pat));

IMPORT ExprPtr bcEval P((IN CONST byteCodeChunk *bc,
			 IN CONST ExprPtr env, ...));
