/* Support functions for compiler to interpreted form.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2021 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "compile.h"
#include "type.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "cmpproto.h"
#include "memproto.h"
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rcompaux.h"
#ifdef HAS_PROTOTYPES
#include "scompaux.h"
#endif

#define __SEG__ COMPILE

#undef DEBUG

EXTERN SHORT generation_count;

PUBLIC COUNTER nLocals(assum)
IN CONST ExprPtr assum;
{
  return C_NLOCALS(assum);
}

PUBLIC void setNLocals(assum,n)
IN CONST ExprPtr assum;
IN COUNTER n;
{
  C_NLOCALS(assum) = n;
}

PRIVATE struct compile_assum *allocCompileAssums=(struct compile_assum *)0L;
PRIVATE struct compile_assum *oldCompileAssums=(struct compile_assum *)0L;

PUBLIC ExprPtr MakeCompAssum(ids,
#if MODULE_SYSTEM & STD_ML_MODULE
			     structs,sigs,
#endif
			     level,ilevel,nlocals,cons,tystate,tytags,imports,
			     fundef,dbgcnt
			     )
IN CONST ExprPtr ids;
#if MODULE_SYSTEM & STD_ML_MODULE
IN CONST ExprPtr structs;
IN CONST ExprPtr sigs;
#endif
IN COUNTER level,ilevel;
IN COUNTER nlocals;
IN CONST ExprPtr cons;
IN CONST ExprPtr tystate;
IN CONST MapPtr tytags;
IN CONST MapPtr imports;
IN CONST ExprPtr fundef;
IN COUNTER dbgcnt;
{
  struct compile_assum *new;

  new = (struct compile_assum *)MALLOC(sizeof(struct compile_assum));
  new->kind = COMPILE_ASSUM_KIND;
  new->mark = 0;
  new->flags = 0;
  /*new->gcmark = 0;*/
  new->next = allocCompileAssums;
  allocCompileAssums = new;
  C_IDS(new) = ids;
#if MODULE_SYSTEM & STD_ML_MODULE
  C_STRUCTS(new) = structs;
  C_SIGS(new) = sigs;
#endif
  C_LEVEL(new) = level;
  C_ILEVEL(new) = ilevel;
  C_NLOCALS(new) = nlocals;
  C_CONS(new) = cons;
  C_TYSTATE(new) = tystate;
  C_TYTAGS(new) = tytags;
  C_IMPORTS(new) = imports;
  C_FUNDEF(new) = fundef;
  C_DBGCNT(new) = dbgcnt;
  return KEEP((ExprPtr)new);
}

PUBLIC ExprPtr NewCompAssum()
{
  return MAKE_TABLE(NewMap(),NewMap(),NewMap(),(COUNTER)0,(COUNTER)0,(COUNTER)0,NewMap(),NewMap(),
		    NewMap(),NIL,NIL,0);
}

PUBLIC ExprPtr DupCompAssum(assum)
IN CONST ExprPtr assum;
{
  ExprPtr new;

  new = MAKE_TABLE(C_IDS(assum),C_STRUCTS(assum),C_SIGS(assum),
		   (COUNTER)C_LEVEL(assum),(COUNTER)C_ILEVEL(assum),
		   (COUNTER)C_NLOCALS(assum),C_CONS(assum),
		   C_TYSTATE(assum),C_TYTAGS(assum),C_IMPORTS(assum),
		   C_FUNDEF(assum),C_DBGCNT(assum));
  C_FLAGS(new) = C_FLAGS(assum);
  return new;
}

PUBLIC void FreeCompileAssums()
{
  struct compile_assum **assump,*assum;

  for (assump = &allocCompileAssums;
       (assum = *assump)!=(struct compile_assum *)0L; )
    {
      if (GC_MARKED(assum))
	{
	  GC_UNMARK(assum);
	  assump = &assum->next;
	}
      else
	{
	  *assump = assum->next;
	  MARKFREE(assum);
	}
    }
}

PUBLIC void FreeCompileAssum(assum)
IN struct compile_assum *assum;
{
  struct compile_assum **assump;

  for (assump = &allocCompileAssums; *assump!=assum; assump = &(*assump)->next);
  /* should not be too slow, since there are few compile assums */
  assum->mark = 0;
  *assump = assum->next;
  MARKFREE(assum);
}

PRIVATE struct compile_assum *CAReserve[MAX_GENERATIONS];

PUBLIC void cutoffInitialCompileAssums()
{
  REGISTER struct compile_assum *assum;

  for (assum = allocCompileAssums; assum!=(struct compile_assum *)0L; assum=assum->next)
    {
      GC_MARK(assum);
    }
  CAReserve[0] = oldCompileAssums = allocCompileAssums;
  allocCompileAssums = (struct compile_assum *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldCompileAssums()
{
  COUNTER i;

  for (i=1; i<generation_count; i++)
    CAReserve[i] = CAReserve[i+1];
}
#endif

PUBLIC void cutoffOldCompileAssums()
{
  REGISTER struct compile_assum **assump,*assum;

  CAReserve[generation_count] = oldCompileAssums;
  for (assump = &allocCompileAssums; (assum = *assump)!=(struct compile_assum *)0L;
       assump = &assum->next)
    GC_MARK(assum);
  *assump = oldCompileAssums;
  oldCompileAssums = allocCompileAssums;
  allocCompileAssums = (struct compile_assum *)0L;
}

PUBLIC void refreshOldCompileAssums()
{
  REGISTER struct compile_assum **assump,*assum,*last;

  last = CAReserve[generation_count];
  for (assump = &oldCompileAssums; assum = *assump, assum!=last; assump = &assum->next)
    {
      GC_UNMARK(assum);
    }
  *assump = allocCompileAssums;
  allocCompileAssums = oldCompileAssums;
  oldCompileAssums = assum;
}

PUBLIC void unmarkOldCompileAssums()
{
  REGISTER struct compile_assum *assum;

  for (assum = oldCompileAssums; assum!=(struct compile_assum *)0L; assum = assum->next)
    GC_UNMARK(assum);
}

PUBLIC MapPtr IdLocs(assum)
IN CONST ExprPtr assum;
{
  return C_IDS(assum);
}

PUBLIC void addIdLocs(id,i_l,assum)
IN CONST ExprPtr id;
IN CONST ExprPtr i_l;
INOUT ExprPtr assum;
{
  setMapPut(id,i_l,&C_IDS(assum));
}

PUBLIC MapPtr TyTags(assum)
IN CONST ExprPtr assum;
{
  return C_TYTAGS(assum);
}

EXTERN EXTENDED zero,fone,ftwo;
PRIVATE EXTENDED failOften=0.10;

PRIVATE ExprPtr optimizePattern P((ExprPtr pat,EXTENDED *w,EXTENDED *failw));

PRIVATE Boolean optReverse(pat1p,pat2p,w,failw)
INOUT ExprPtr *pat1p,*pat2p;
OUT EXTENDED *w,*failw;
{
  EXTENDED w1,w2,failw1,failw2,s1,s2;
  Boolean revp;

  *pat1p = optimizePattern(*pat1p,&w1,&failw1);
  *pat2p = optimizePattern(*pat2p,&w2,&failw2);
  /* We evaluate heuristically the one of pat1, pat2 that has
  	the smallest failw/w.
     We estimate *failw to be (failw1/w1+failw2/w2)*(w1+w2+1)/2 if pat1 fails faster. */
  *w = w1+w2+fone;
  s1 = failw1/w1;
  s2 = failw2/w2;
  revp = (s1>s2);
  *failw = ldexp((s1+s2)*(*w),-1);
  return revp;
}

#ifdef OBSOLETE
	/* former optimizations for tuples and records as vectors.
	   The formulas were so neat that I didn't have the courage
	   to erase the code altogether. */

PRIVATE int optswpcmp(swp1,swp2)
IN CONST ANSI_CONST struct optswp *swp1,*swp2;
{
  EXTENDED s1,s2;

  s1 = swp1->failw/swp1->w;
  s2 = swp2->failw/swp2->w;
  if (s1<s2)
    return -1;
  if (s1>s2)
    return 1;
  return 0;
}

PRIVATE void computeW(swps,n,w,failw)
IN CONST struct optswp *swps;
IN COUNTER n;
OUT EXTENDED *w,*failw;
{
  COUNTER i;
  struct optswp *swpp;

  *w = fone;
  *failw = zero;
  /* w=w1+w2+...+wn+1, failw=(failw1/w1+...+failwn/wn)*w/n */
  for (i=0, swpp=swps; i<n; i++, swpp++)
    {
      *failw += swpp->failw/swpp->w;
      *w += swpp->w;
    }
  *failw *= (*w)/(EXTENDED)n;
}

PRIVATE ExprPtr optSwpTuple(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  COUNTER i,n;
  EXTENDED failw0;
  ExprPtr indstr;
  INT size;

  n = AVAL(pat)->n;
  {
    struct optswp *swpp;
    struct optswp *swps;

    swps = (struct optswp *)alloctmp_anchored(n*sizeof(struct optswp), (char *)&pat);
    for (i=0, swpp=swps; i<n; i++, swpp++)
      {
	swpp->index.i = i;
	optimizePattern(AVAL(pat)->expr[i],&swpp->w,&swpp->failw);
      }
    computeW(swps,n,w,&failw0);
    qsort(swps,(long)n,(long)sizeof(struct optswp),optswpcmp);
    /* as a heuristic, we order the patterns so that the ones
       that fail the fastest are put at the beginning */
    computeW(swps,n,w,failw);
    if ((*failw)+(EXTENDED)n<=failw0) /* if swaps make it so quick that it offsets the cost of indirection */
      {
	size = ((long)n)*sizeof(COUNTER);
	{
	  COUNTER *ip;
#if STOP_AND_COPY_VECTORS
	  COUNTER *indexes;

	  indexes = (COUNTER *)alloctmp_anchored(size, (char *)&pat);
	  for (i=0, swpp=swps, ip=indexes; i<n; i++)
	    *ip++ = (swpp++)->index.i;
	  indstr = Kstring1(size,(CHAR *)indexes,true);
#else
	  COUNTER *indexes;

	  VECALLOC(indexes,size);
	  for (i=0, swpp=swps, ip=indexes; i<n; i++)
	    *ip++ = (swpp++)->index.i;
	  indstr = Kstring1(size,(CHAR *)indexes,false);
#endif
	}
      }
    else
      indstr = NIL;
  }
  return indstr;
}

PRIVATE void optimizeTupleGnrl(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  ExprPtr indstr;

  indstr = optSwpTuple(pat,w,failw);
  if (indstr!=NIL)
  {
    if (pat->mark & MARK_SHARED)
      berror(message(69));
    SETSHAPE(pat,indstr);
    pat->kind = E_PAT_SWPTUPLE;
  }
}

MAPSWEEPFUNC(do_setswprec)
{
  struct swprec *data;

  data = MAPDATA(struct swprec *);
  data->swpp->index.lbl = x;
  optimizePattern(AVAL(data->pat)->expr[IVAL(y)-1],
		  &data->swpp->w,&data->swpp->failw);
  data->swpp++;
  return false;
}

PRIVATE ExprPtr optSwpRecord(shape,pat,w,failw)
IN CONST MapPtr shape;
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  COUNTER i,n;
  struct swprec data;
  EXTENDED failw0;
  VecPtr lbls;
  ExprPtr *ep;
  ExprPtr env;
  long size;

  n = AVAL(pat)->n;
  data.pat = pat;
  {
    data.swps = (struct optswp *)alloctmp(n*sizeof(struct optswp));
    data.swpp = data.swps;
    (void) MapSweep(shape,do_setswprec,&data);
    computeW(data.swps,n,w,&failw0);
    qsort(data.swps,(long)n,(long)sizeof(struct optswp),optswpcmp);
    /* as a heuristic, we order the patterns so that the ones
       that fail the fastest are put at the beginning */
    computeW(data.swps,n,w,failw);
    if ((*failw)+(EXTENDED)n<=failw0) /* if swaps make it so quick that it offsets the cost of indirection */
      {
	size = VECSIZE(n);
	CHECKHETEROFREE();
	scAlloc(*(char **)&lbls,size);
	lbls->n = n;
	for (i=0, data.swpp=data.swps, ep=lbls->expr; i<n; i++)
	  *ep++ = (data.swpp++)->index.lbl;
	MAKEENV_NOGC(lbls,shape,env);
      }
    else
      env = NIL;
  }
  return env;
}

PRIVATE void optimizeRecGnrl(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  ExprPtr indstr,shape,lbl;

  if (pat->mark & MARK_SHARED)
    berror(message(69));
  switch (pat->kind)
  {
    case E_PAT_SWPRECOPT:
      SETSHAPE(pat,CAR(ASHAPE(pat)));
      break;
    case E_PAT_SWPRECORD:
      SETSHAPE(pat,ENEXT(ASHAPE(pat)));
      break;
  }
  lbl = ChooseInDom(ASHAPE(pat));
  if (lbl->kind==(CHAR)E_INT) /* precompiled record pattern */
    {
      indstr = optSwpTuple(pat,w,failw);
      if (indstr!=NIL)
	{
	  MAKECONS(ASHAPE(pat),indstr,shape);
	  (void) KEEP(shape);
	  SETSHAPE(pat,shape);
	  pat->kind = E_PAT_SWPRECOPT;
	}
    }
  else /* extensible record pattern */
    {
      shape = optSwpRecord(ASHAPE(pat),pat,w,failw);
      if (shape!=NIL)
	{
	  SETSHAPE(pat,shape);
	  pat->kind = E_PAT_SWPRECORD;
	}
    }
}
#endif

PRIVATE void typeWeight P((IN CONST ExprPtr type,INOUT EXTENDED *w,INOUT EXTENDED *failw));

MAPSWEEPFUNC(do_typeweight)
{
  struct typeweight *data;
  EXTENDED w,failw;

  data = MAPDATA(struct typeweight *);
  typeWeight(y,&w,&failw);
  w += fone;
  failw += fone;
  data->sumw += w;
  data->prodw *= w;
  data->prodfailw *= failw;
  return false;
}

MAPSWEEPFUNC(do_typeweight_num)
{
  struct typeweight *data;
  EXTENDED w,failw;

  data = MAPDATA(struct typeweight *);
  typeWeight(x,&w,&failw);
  w += fone;
  failw += failOften;
  data->sumw += w;
  data->prodw *= w;
  data->prodfailw *= failw;
  return false;
}

DEF_PRIVATE_ACTION_1ARG(typeweightk)
{
  struct typeweightk *data;

  data = (struct typeweightk *)ENV_UNDERLYING(env);
  typeWeight(data->type,&data->w,&data->failw);
  return NIL;
}
ACTION_END

PRIVATE void typeWeightk(type,w,failw)
IN CONST ExprPtr type;
INOUT EXTENDED *w,*failw;
{
  struct typeweightk *data = (struct typeweightk *)alloctmp(sizeof(struct typeweightk));

  data->type = type;
  data->w = *w;
  data->failw = *failw;
  (void) CallCC(ACTION_NAME(typeweightk),(ExprPtr)data,true);
  *w = data->w;
  *failw = data->failw;
}

PRIVATE void typeWeight(type,w,failw)
IN CONST ExprPtr type;
INOUT EXTENDED *w,*failw;
{
  if (STKFULL(type))
    {
      typeWeightk(type,w,failw);
      return;
    }
  /*if (type==NIL)
  {
    *w = fone;
    *failw = failOften;
  }
  else*/
  switch (type->kind)
  {
    case E_NIL:
      *w = fone;
      *failw = failOften;
      break;
    case E_VAR:
      *w = ftwo;
      *failw = fhalf;
      break;
    case E_TYPE_APPL:
      typeWeight(APPLARG(type),w,failw); /* like E_CONS with w1=1, failw1=failOften */
      *failw = (((*w)+fone)*failOften*(*failw))/(*w);
      *w += fone;
      break;
    case E_CONS: /* w/failw=w1/failw1.w2/failw2, w=w1+w2 */
      {
	EXTENDED w1,failw1,w2,failw2;

	typeWeight(CAR(type),&w1,&failw1);
	typeWeight(CDR(type),&w2,&failw2);
	*w = w1+w2;
	*failw = ((*w)*failw1*failw2)/(w1*w2);
	break;
      }
    case E_EMPTY:
    case E_MAPLET:
    case E_MAP:
      {
	struct typeweight data;

	data.sumw = zero;
	data.prodw = fone;
	data.prodfailw = fone;
	(void) MapSweep(type,do_typeweight,&data);
	*w = fone+data.sumw;
	*failw = failOften+(data.sumw*data.prodw)/data.prodfailw;
	break;
      }
    case E_NUMTYPE:
      {
	struct typeweight data;

	data.sumw = zero;
	data.prodw = fone;
	data.prodfailw = fone;
	(void) MapSweep(REFVAL(type),do_typeweight_num,&data);
	*w = fone+data.sumw;
	*failw = failOften+(data.sumw*data.prodw)/data.prodfailw;
	break;
      }
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    case E_TYPEFX: /* fallthrough */
#endif
    default:
      kinderror(message(456),type->kind);
      break;
  }
}

PRIVATE void addTypeWeight(type,w,failw)
IN CONST ExprPtr type;
INOUT EXTENDED *w,*failw;
{
  EXTENDED ww,failww;

  typeWeight(type,&ww,&failww);
  *w += ww;
  *failw += failww;
}

PRIVATE int optmapaddcmp(_swp1,_swp2)
IN CONST ANSI_CONST VOIDPTR _swp1;
IN CONST ANSI_CONST VOIDPTR _swp2;
{
  REGISTER struct optmapadd *swp1 = (struct optmapadd *)_swp1;
  REGISTER struct optmapadd *swp2 = (struct optmapadd *)_swp2;

  if (swp1->failw<swp2->failw)
    return -1;
  if (swp1->failw>swp2->failw)
    return 1;
  return 0;
}

#ifdef INCORRECT
	/* optimizing mapadds this way, by rearranging the order of
	   comparison of patterns, is incorrect. Indeed, the semantics
	   of the pattern {_,x,...} is to get the second element of
	   a set, but it could be optimized here into
	   {x,_,...} (with condition card>=2), which would get the
	   first element of the set.
	*/

PRIVATE ExprPtr optimizePatMapAdd(pat,w,failw,cardpat)
IN CONST ExprPtr pat,cardpat;
OUT EXTENDED *w,*failw;
{
  ExprPtr pp;
  EXTENDED wr,failwr,nn;
  COUNTER i,n;
  Boolean exact;

  /* among x and y, the first to be evaluated will be the one
  	with the smallest w.
    If the costs of matching the (x,y) are wi, failwi (i=1..m)
    and the set has cardinal n>=m, the costs of matching the
    complete list of mapadds is:
    w = n(w1+2+(n-1)(w2+2+...+(n-m+1)(wm+2+wrest)...)+failw2)+failw1 at worst
    and
    failw = n(w1+2+(n-1)(w2+2+...+(n-m+1)(wm+2+failwrest)...)+failw2)+failw1
    	On average, we multiply the n coefficients by 1/2.
    Then we order the list of mapadds by putting those that
    fail the fastest (failw1 minimum)
 */
  for (n=0, pp=pat; /*pp!=NIL && */pp->kind==(CHAR)E_PAT_MAPADD; n++, pp=PATMAPADDREST(pp));
  {
    struct optmapadd *swpp;
    struct optmapadd *swps;

    swps = (struct optmapadd *)alloctmp((long)(n*sizeof(struct optmapadd)));
    for (i=0, swpp=swps, pp=pat; i<n; i++, pp=PATMAPADDREST(pp), swpp++)
      {
	swpp->xy = optimizePattern(PATMAPADDXY(pp),&swpp->w,&swpp->failw);
      }
    qsort(swps,(long)n,(long)sizeof(struct optmapadd),optmapaddcmp);
    pp = optimizePattern(pp,&wr,&failwr); /* optimize the rest */
    if (exact = (pp==NewMap()))
      nn = zero;
    else
      nn = fone; /* estimated size of rest, times 2 */
    *w = wr;
    *failw = failwr;
    for (i=n, swpp = &swps[i]; nn += fhalf, --swpp, --i>=0; )
      {
	*w = nn*((*w)+ftwo+swpp->w)+swpp->failw;
	*failw = nn*((*failw)+ftwo+swpp->w)+swpp->failw;
      }
    for (i=0, swpp=swps, pp=pat; i<n; i++, swpp++, pp=PATMAPADDREST(pp))
      PATMAPADDXY(pp) = swpp->xy; /* permute elements */
    if (cardpat==pat && (exact || n>=2)) /* if no cardinal testing pattern was compiled,
					    and if it is worth making one, then we hack
					    one just after the first PAT_MAPADD,
					    and we exchange it with the first one, so
					    that the address of the whole pattern
					    does not change. */
      {
	pat = MakeCounted((exact?E_PAT_CARD_EQ:E_PAT_CARD_GEQ),pat,(INT)n);
	(void) KEEP(pat);
      }
  }
  return pat;
}
#endif

PRIVATE ExprPtr optimizePatMapAdd(pat,w,failw,cardpat)
IN CONST ExprPtr pat,cardpat;
OUT EXTENDED *w,*failw;
{
  ExprPtr pp,rest;
  EXTENDED wr,failwr,nn;
  COUNTER i,endi,n;
#if !TRANSLATOR
  Boolean exact,ellipsis;
#endif

  /* If the costs of matching the (x,y) are wi, failwi (i=1..m)
    and the set has cardinal n>=m, the costs of matching the
    complete list of mapadds is:
    w = n(w1+2+(n-1)(w2+2+...+(n-m+1)(wm+2+wrest)...)+failw2)+failw1 at worst
    and
    failw = n(w1+2+(n-1)(w2+2+...+(n-m+1)(wm+2+failwrest)...)+failw2)+failw1
    	On average, we multiply the n coefficients by 1/2.
    Then we don't reorder the list of mapadds, because it would be
    incorrect. However, we eliminate final underscores the following way:
    {a => b,_,...} becomes {a => b,...}, with a cardinality constraint.
 */
  for (n=0, pp=pat; /*pp!=NIL && */pp->kind==(CHAR)E_PAT_MAPADD; n++, pp=PATMAPADDREST(pp))
    ;
  {
    struct optmapadd *swpp;
    struct optmapadd *swps;

    swps = (struct optmapadd *)alloctmp((long)(n*sizeof(struct optmapadd)));
    for (i=0, pp=pat, swpp = swps; i<n; i++, pp=PATMAPADDREST(pp), swpp++)
      {
	swpp->xy = optimizePattern(PATMAPADDXY(pp),&swpp->w,&swpp->failw);
      }
    rest = optimizePattern(pp,&wr,&failwr); /* optimize the rest */
#if TRANSLATOR
    if (rest==NewMap())
      nn = zero;
#else
    if ((exact = (rest==NewMap())))
      nn = zero;
#endif
    else
      {
#if !TRANSLATOR
	ellipsis = (rest==patany);
#endif
	nn = fone; /* estimated size of rest, times 2 */
      }
    *w = wr;
    *failw = failwr;
    endi = n;
#if !TRANSLATOR
    if (exact || ellipsis) /* in the case of {something} or {something,...}
			      patterns, we drop the trailing underscores and
			      add a cardinal checking pattern. */
      {
	for (swpp = &swps[endi]; --swpp, --endi>=0; )
	  {
	    if (swpp->xy!=patany)
	      break;
	  }
	endi++;
      }
#endif
    /* now, compute cost */
    for (i=endi, swpp = &swps[i]; nn += fhalf, --swpp, --i>=0; )
      {
	*w = nn*((*w)+ftwo+swpp->w)+swpp->failw;
	*failw = nn*((*failw)+ftwo+swpp->w)+swpp->failw;
      }
    /* install all corrected patterns back in place */
    for (i=0, swpp=swps, pp=pat; i<endi; i++, swpp++, pp=PATMAPADDREST(pp))
      PATMAPADDXY(pp) = swpp->xy;
#if !TRANSLATOR
    /* install the rest, when there were underscores at the end */
    if (endi!=n)
      {
	if (endi==0)
	  pat = rest;
	else
	  {
	    for (i=1, pp=pat; i<endi; pp=PATMAPADDREST(pp));
	    SETQ(PATMAPADDREST(pp),rest);
	    HITBARRIER(pp);
	  }
      }
    if (cardpat==pat && (exact || ellipsis || n>=2/* || endi!=n*/))
      /* if no cardinal testing pattern was compiled,
	 and if it is worth making one, then we hack one just after the first PAT_MAPADD,
	 and we exchange it with the first one, so that the address of the whole pattern
	 does not change. */
      {
	pat = MakeCounted((exact?E_PAT_CARD_EQ:E_PAT_CARD_GEQ),pat,(INT)n);
	(void) KEEP(pat);
      }
#endif
  }
  return pat;
}

PRIVATE ExprPtr optimizePatRecMapAdd(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  ExprPtr pp;
  EXTENDED wr,failwr,nn;
  COUNTER i,n;

  /* We can reorder record mapadds. */
  for (n=0, pp=pat; /*pp!=NIL && */pp->kind==(CHAR)E_PAT_RECORD_MAPADD; n++, pp=CDR(pp))
    ;
  {
    struct optmapadd *swpp;
    struct optmapadd *swps;

    swps = (struct optmapadd *)alloctmp((long)(n*sizeof(struct optmapadd)));
    for (i=0, swpp=swps, pp=pat; i<n; i++, pp=CDR(pp), swpp++)
      {
	swpp->xy = optimizePattern(CAR(pp),&swpp->w,&swpp->failw);
      }
    qsort(swps,(long)n,(long)sizeof(struct optmapadd),optmapaddcmp);
    (void) optimizePattern(pp,&wr,&failwr); /* optimize the rest */
    nn = zero; /* estimated size of rest, times 2 */
    *w = wr;
    *failw = failwr;
    for (i=n, swpp = &swps[i]; nn += fhalf, --swpp, --i>=0; )
      {
	*w = nn*((*w)+ftwo+swpp->w)+swpp->failw;
	*failw = nn*((*failw)+ftwo+swpp->w)+swpp->failw;
      }
    for (i=0, swpp=swps, pp=pat; i<n; i++, swpp++, pp=CDR(pp))
      CAR(pp) = swpp->xy; /* permute elements */
  }
  return pat;
}

PRIVATE ExprPtr optPatCouple(pat,w,failw,kind,revkind)
INOUT ExprPtr pat;
OUT EXTENDED *w,*failw;
IN ExprKind kind,revkind;
{
  ExprPtr car,cdr;

  car = CAR(pat);
  cdr = CDR(pat);
  pat->kind = (CHAR)((optReverse(&car,&cdr,w,failw))?revkind:kind);
  if (car!=CAR(pat))
    SETCAR(pat,car);
  if (cdr!=CDR(pat))
    SETCDR(pat,cdr);
  return pat;
}

PRIVATE ExprPtr optPair(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  ExprPtr car,cdr;

  car = CAR(pat);
  cdr = CDR(pat);
  (void) optReverse(&cdr,&car,w,failw);
  if (car!=CAR(pat))
    SETCAR(pat,car);
  if (cdr!=CDR(pat))
    SETCDR(pat,cdr);
  return pat;
}
#define optimizePatAs optPair

#ifdef OBSOLETE
PRIVATE ExprPtr optimizePatAs(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  ExprPtr car,cdr;

  car = PATAS1(pat);
  cdr = PATAS2(pat);
  if (optReverse(&cdr,&car,w,failw))
  {
    if (car!=CDR(pat))
      SETCDR(pat,car);
    if (cdr!=CAR(pat))
      SETCAR(pat,cdr);
  }
  else
  {
    if (car!=CAR(pat))
      SETCAR(pat,car);
    if (cdr!=CDR(pat))
      SETCDR(pat,cdr);
  }
  return pat;
}
#endif

DEF_PRIVATE_ACTION_1ARG(optpatk)
{
  struct optpatk *data;

  data = (struct optpatk *)ENV_UNDERLYING(env);
  return optimizePattern(data->pat,&data->w,&data->failw);
}
ACTION_END

PRIVATE ExprPtr optimizePatternk(pat,w,failw)
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  struct optpatk *data = (struct optpatk *)alloctmp(sizeof(struct optpatk));
  ExprPtr res;

  data->pat = pat;
  data->w = *w;
  data->failw = *failw;
  res = CallCC(ACTION_NAME(optpatk),(ExprPtr)data,true);
  *w = data->w;
  *failw = data->failw;
  return res;
}

PRIVATE ExprPtr optimizePattern(pat,w,failw)
 /* the w is an estimation of the cost of succesful matching,
	failw<=w is an estimation of the cost of a failing match.
	Returns new (modified) pattern. */
IN CONST ExprPtr pat;
OUT EXTENDED *w,*failw;
{
  if (STKFULL(pat))
    return optimizePatternk(pat,w,failw);
  /*if (pat==NIL)
    {
      *w = fone;
      *failw = fbig;
    }
  else*/
  switch (pat->kind)
    {
    case E_VARINTRO:
      *w = ftwo;
      *failw = ftwo; /* never fails */
      break;
    case E_PAT_ANY:
    case E_TUPLE_0: /* tuple0 and record0 never fail, because of typing. */
    case E_RECORD_0:
      *w = fone;
      *failw = fone; /* never fails */
      break;
    case E_NIL:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_VARREF:
      *w = fone;
      *failw = failOften; /* has a good chance of failing */
      break;
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_REF: case E_PAT_CONS:
#ifdef DATACONPAT
    case E_DATACON:
#endif
      {
	ExprPtr aux;

	aux = optimizePattern(REFVAL(pat),w,failw);
	if (aux!=REFVAL(pat))
	  SETREFVAL(pat,aux);
	*w += fone;
	*failw += fone; /* if it fails, it is because of the REFVAL, not of the type */
	break;
      }
    case E_MAPLET:
      pat = optPair(pat,w,failw);
      break;
    case E_DATA:
      {
	ExprPtr aux;

	aux = optimizePattern(CDR(pat),w,failw);
	if (aux!=CDR(pat))
	  SETCDR(pat,aux);
	/* Let w1 be the current *w, failw1 be the current *failw,
	   and n be the number of constructors in the datatype.
	   We have w=w1+1.  To estimate failw, we want to
	   have:
	   - if n=1, then failw=failw1+1;
	   - when n->infty, failw->0.
	   We take failw = (failw1+1)/n.
	   */
	*w += fone;
	*failw = (*failw + fone)/((EXTENDED)AVAL(CAR(CAR(pat)))->n);
	break;
      }
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      {
	ExprPtr aux;

	aux = optimizePattern(CDR(pat),w,failw);
	if (aux!=CDR(pat))
	  SETCDR(pat,aux);
	*w += fone;
	*failw += fone; /* if it fails, it is because of the REFVAL, not of the type */
	break;
      }
    case E_CONS:
    case E_UNSHARED_CONS:
    case E_PAT_REVCONS:
      pat = optPatCouple(pat,w,failw,E_CONS,E_PAT_REVCONS);
      break;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_REVTUPLE_2:
      pat = optPatCouple(pat,w,failw,E_TUPLE_2,E_PAT_REVTUPLE_2);
      break;
    case E_APPL:
      pat = optPatCouple(pat,w,failw,E_APPL,E_APPL);
      break;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_GNRL_CONS:
      pat = optPatCouple(pat,w,failw,E_PAT_TUPLE_GNRL_CONS,E_PAT_REVTUPLE_GNRL_CONS);
      break;
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      pat = optPatCouple(pat,w,failw,E_PAT_RECORD_GNRL_CONS,E_PAT_REVRECORD_GNRL_CONS);
      break;
    case E_PAT_AS:
      pat = optimizePatAs(pat,w,failw);
      break;
    case E_PAT_CHECK:
      {
	ExprPtr aux;

	aux = optimizePattern(PATCHECKVAL(pat),w,failw);
	if (aux!=PATCHECKVAL(pat))
	  {
	    SETQ(PATCHECKVAL(pat),aux);
	    HITBARRIER(pat);
	  }
	*w += fone;
	*failw += fhalf; /* if it fails, it may be because of the tag */
	break;
      }
    case E_PAT_PACK:
      {
	ExprPtr aux;

	aux = optimizePattern(PACKVAL(pat),w,failw);
	if (aux!=PACKVAL(pat))
	  {
	    SETQ(PACKVAL(pat),aux);
	    HITBARRIER(pat);
	  }
	addTypeWeight(PACKTYPE(pat),w,failw);
	break;
      }
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
      {
	ExprPtr aux;

	aux = optimizePatMapAdd(COUNTEDVAL(pat),w,failw,pat);
	if (aux!=COUNTEDVAL(pat))
	  SETREFVAL(pat,aux);
	break;
      }
    case E_PAT_MAPADD:
      pat = optimizePatMapAdd(pat,w,failw,pat);
      break;
    case E_PAT_RECORD_MAPADD:
      pat = optimizePatRecMapAdd(pat,w,failw);
      break;
    case E_DEBUGNAME:
      {
	ExprPtr aux;

	aux = optimizePattern(DNPAT(pat),w,failw);
	if (aux!=DNPAT(pat))
	  {
	    SETQ(DNPAT(pat),aux);
	    HITBARRIER(pat);
	  }
#ifdef OLDDEBUG
	*w += fone;
	*failw += fone;
#endif
	break;
      }
    default:
      kinderror(message(444),pat->kind);
  }
  return pat;
}

PUBLIC ExprPtr optPattern(pat)
INOUT ExprPtr pat;
{
  EXTENDED w,failw;

  return optimizePattern(pat,&w,&failw);
}

/* now that we know how to optimize single patterns, we shall
   organize matches, i.e. lists of (pattern,body) pairs.
	The idea is to share common prefixes of patterns as
   much as we can, then to insert an E_PATCHOICE when the patterns
   first differ. Variable indexes are renamed so as to share more.
	Notice that usually, in debug mode, we use the fact that
   more recently bound variables have higher indexes (like in
   val a=...;val a=..., the a in scope is the last one), but as
   patterns are linear in matches, this does not cause any harm.
        Still in debug mode, when we rename variable indexes,
   we insert a few DEBUGNAMEs to make MatchValue() record the change
   in names at each PATCHOICE point. For example, in fn (x,y,0) => x | (y,x,z) => y,
   we record that on the second branch, slot 0 is now "y" and slot 1 "x".
*/

PRIVATE permutationPtr nilPermutation()
{
  ExprPtr e;

  MAKECONS(NIL,NIL,e);
  (void) KEEP(e);
  SETCDR(e,e);
  return e;
}

#define nilPp(p) (CAR(p)==NIL)

PRIVATE COUNTER applyPermutation(p,i)
IN CONST permutationPtr p;
IN COUNTER i;
{
  ExprPtr swp;

  p = CAR(p);
  while (p!=NIL)
    {
      swp = CAR(p);
      p = CDR(p);
      if (i==VARRANK(swp))
	i = VARLEVEL(swp);
      else if (i==VARLEVEL(swp))
	i = VARRANK(swp);
    }
  return i;
}

PRIVATE void extendPermutation(p,i1,i2)
INOUT permutationPtr p;
IN COUNTER i1,i2;
{
  ExprPtr var,e,last;
  ENTER;

  if (i1!=i2)
    {
      var = Kvarref(i1,i2);
      (void) KEEP(var);
      MAKECONS(var,NIL,e);
      last = CDR(p);
      if (last==p)
	{
	  SETCAR(p,e);
	  SETCDR(p,e);
	}
      else
	{
	  SETCDR(last,e);
	}
    }
  QUIT;
}

PRIVATE ExprPtr renameVarref(var,p,level,shift)
IN CONST ExprPtr var;
IN CONST permutationPtr p;
IN COUNTER level,shift; /* shift should be >=0 */
{
  COUNTER i1,i2;

  if (level<=VARLEVEL(var))
    {
      i1 = VARRANK(var);
      if (level==VARLEVEL(var))
	i2 = applyPermutation(p,i1);
      else
	i2 = i1;
      if (i1!=i2 || shift!=0)
	{
	  var = Kvarref(i2,VARLEVEL(var)+shift);
	  (void) KEEP(var);
	}
    }
  return var;
}

PRIVATE ExprPtr MakePatCheck(val,bits)
IN CONST ExprPtr val;
IN CONST CHAR *bits;
{
  char dum[(E_UNUSED+7) >> 3];
  INT n;
  ExprPtr e;

  n = (E_UNUSED+7) >> 3;
  safememcpy(dum,bits,n);
  CHECKHETEROFREE();
  scAlloc(bits,n);
  safememcpy(bits,dum,n);
  EALLOC_HETERO_NOGC(e);
  e->kind = (CHAR)E_PAT_CHECK;
  PATCHECKVAL(e) = val;
  PATCHECKBITS(e) = bits;
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr renameExpr P((IN CONST ExprPtr expr,IN CONST permutationPtr p,
			      IN COUNTER ilevel,IN COUNTER level,IN COUNTER shift));
PRIVATE ExprPtr renamePattern P((IN CONST ExprPtr pat,IN CONST permutationPtr p,
				 IN COUNTER ilevel,IN COUNTER level,IN COUNTER shift,
				 INOUT ExprPtr dbgnames));

DEF_PRIVATE_ACTION_1ARG(renpatk)
{
  struct renpatk *data;

  data = (struct renpatk *)ENV_UNDERLYING(env);
  return renamePattern(data->pat,data->p,
		       data->ilevel,data->level,data->shift,data->dbgnames);
}
ACTION_END

PRIVATE ExprPtr renamePatternk(pat,p,ilevel,level,shift,dbgnames)
IN CONST ExprPtr pat;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
INOUT ExprPtr dbgnames;
{
  struct renpatk *data = (struct renpatk *)alloctmp(sizeof(struct renpatk));

  data->pat = pat;
  data->p = p;
  data->ilevel = ilevel;
  data->level = level;
  data->shift = shift;
  data->dbgnames = dbgnames;
  return CallCC(ACTION_NAME(renpatk),(ExprPtr)data,true);
}

PRIVATE ExprPtr renamePattern(pat,p,ilevel,level,shift,dbgnames)
IN CONST ExprPtr pat;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
INOUT ExprPtr dbgnames;
{
  ExprPtr aux;

  if (STKFULL(pat))
    return renamePatternk(pat,p,ilevel,level,shift,dbgnames);
  /*if (pat==NIL)
    return false;*/
  switch (pat->kind)
    {
    case E_VARINTRO:
      if (ilevel==0)
	{
	  COUNTER i1,i2;

	  i1 = VARRANK(pat);
	  i2 = applyPermutation(p,i1);
	  if (i1!=i2)
	    {
	      pat = Kvarintro(i2);
	      (void) KEEP(pat);
	    }
	}
	break;
    case E_DEBUGNAME:
      if (ilevel==0)
	{
	  COUNTER i1;

	  i1 = COUNTEDN(DNINFO(pat))
#ifdef OLDDEBUG
	    -1
#endif
	      ;
	  i1 = applyPermutation(p,i1)
#ifdef OLDDEBUG
	    +1
#endif
	      ;
	  SETCOUNTEDN(DNINFO(pat),i1);
	  if (dbgnames!=NIL)
	    setrefMapPut(pat,UNIT,dbgnames);
	}
      aux = renamePattern(DNPAT(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=DNPAT(pat))
	{
	  pat = MakeDebugName(aux,DNINFO(pat));
	  (void) KEEP(pat);
	}
      break;
    case E_PATCHOICE:
      {
	ExprPtr match,car,pat2;

	car = renamePattern(CAR(pat),p,ilevel,level,shift,dbgnames);
	match = CDR(pat);
	pat2 = renamePattern(CAR(match),p,
			     ilevel,level,shift,dbgnames);
	aux = renameExpr(CDR(match),p,ilevel,ilevel,shift);
	/* notice that the level to submit to renameExpr() is actually ilevel
	   (this is not a typo). */
	if (pat2!=CAR(match) || aux!=CDR(match))
	  {
	    aux = MakePair((ExprKind)match->kind,pat2,aux);
	    (void) KEEP(aux);
	  }
	else aux = match;
	if (car!=CAR(pat) || aux!=match)
	  {
	    pat = MakePair(E_PATCHOICE,car,aux);
	    (void) KEEP(pat);
	  }
      }
      break;
    case E_CONS: case E_UNSHARED_CONS: case E_PAT_REVCONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2: case E_PAT_REVTUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_AS: case E_MAPLET:
    case E_PAT_MAPADD: case E_PAT_RECORD_MAPADD:
    case E_APPL:
      {
	ExprPtr car;

	car = renamePattern(CAR(pat),p,
			    ilevel,level,shift,dbgnames);
	aux = renamePattern(CDR(pat),p,ilevel,level,shift,dbgnames);
	if (car!=CAR(pat) || aux!=CDR(pat))
	  {
	    pat = MakePair((ExprKind)pat->kind,car,aux);
	    (void) KEEP(pat);
	  }
      }
      break;
    case E_NIL:
    case E_RECORD_0: case E_TUPLE_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE:
    case E_PAT_ANY:
      break;
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
    case E_REF:
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
    case E_PAT_CONS:
#ifdef DATACONPAT
    case E_DATACON:
#endif
      aux = renamePattern(REFVAL(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=REFVAL(pat))
	{
	  pat = MakeCounted((ExprKind)pat->kind,aux,COUNTEDN(pat));
	  (void) KEEP(pat);
	}
      break;
    case E_DATA:
      aux = renamePattern(CDR(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=CDR(pat))
	{
	  pat = MakeData(pat->magic,CAR(pat),aux);
	  (void) KEEP(pat);
	}
      break;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      aux = renamePattern(CDR(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=CDR(pat))
	{
	  pat = MakePair((ExprKind)pat->kind,CAR(pat),aux);
	  (void) KEEP(pat);
	}
      break;
    case E_VARREF:
      return renameVarref(pat,p,level,shift);
    case E_PAT_CHECK:
      aux =  renamePattern(PATCHECKVAL(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=PATCHECKVAL(pat))
	{
	  pat = MakePatCheck(aux,PATCHECKBITS(pat));
	}
      break;
    case E_PAT_PACK:
      aux = renamePattern(PACKVAL(pat),p,ilevel,level,shift,dbgnames);
      if (aux!=PACKVAL(pat))
	{
	  pat = MakePack(E_PAT_PACK,aux,PACKTYPE(pat));
	  (void) KEEP(pat);
	}
      break;
    default:
      kinderror(message(634),pat->kind);
      break;
    }
  return pat;
}

PRIVATE ExprPtr renameBindings P((IN CONST ExprPtr bindings,IN CONST permutationPtr p,
				  IN COUNTER ilevel,IN COUNTER level,IN COUNTER shift));

DEF_PRIVATE_ACTION_1ARG(renbindk)
{
  struct renexprk *data;

  data = (struct renexprk *)ENV_UNDERLYING(env);
  return renameBindings(data->expr,data->p,data->ilevel,data->level,data->shift);
}
ACTION_END

PRIVATE ExprPtr renameBindingsk(bindings,p,ilevel,level,shift)
IN CONST ExprPtr bindings;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  struct renexprk *data = (struct renexprk *)alloctmp(sizeof(struct renexprk));

  data->expr = bindings;
  data->p = p;
  data->ilevel = ilevel;
  data->level = level;
  data->shift = shift;
  return CallCC(ACTION_NAME(renbindk),(ExprPtr)data,true);
}

PRIVATE ExprPtr renameBindings(bindings,p,ilevel,level,shift)
INOUT ExprPtr bindings;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  ExprPtr binding;

  if (bindings!=NIL)
    {
      if (STKFULL(bindings))
	return renameBindingsk(bindings,p,ilevel,level,shift);
      binding = CAR(bindings);
      switch (binding->kind)
	{
	case E_LET: /* local bindings */
	  {
	    COUNTER newlevel = level+1;
	    ExprPtr bind,body;

	    if (IVAL(CAR(LETBIND(binding)))<0)
	      newlevel++;
	    bind = renameBindings(CDR(LETBIND(binding)),p,newlevel,
				  newlevel,shift);
	    body = renameBindings(LETBODY(binding),p,ilevel,newlevel,shift);
	    if (body!=LETBODY(binding) || bind!=CDR(LETBIND(binding)))
	      {
		bind = MakePair(E_CONS,CAR(LETBIND(binding)),bind);
		(void) KEEP(bind);
		binding = MakeLet(bind,body);
		(void) KEEP(binding);
	      }
	    break;
	  }
	case E_DEBUGDECL:
	  binding = CAR(bindings);
	  break;
	default:
	  {
	    ExprPtr pat,expr;

	    if (shift!=0)
	      pat = renamePattern(CAR(binding),p,ilevel,level,shift,NIL);
	    else pat = CAR(binding); /* if shift==0, then the pattern cannot change. */
	    expr = renameExpr(CDR(binding),p,ilevel,level,shift);
	    if (pat!=CAR(binding) || expr!=CDR(binding))
	      {
		binding = MakePair(E_CONS,pat,expr);
		(void) KEEP(binding);
	      }
	    break;
	  }
	}
      {
	ExprPtr rest;

	rest = renameBindings(CDR(bindings),p,ilevel,level,shift);
	if (binding!=CAR(bindings) || rest!=CDR(bindings))
	  {
	    bindings = MakePair(E_CONS,binding,rest);
	    (void) KEEP(bindings);
	  }
      }
    }
  return bindings;
}

PRIVATE ExprPtr renameDomains P((IN CONST ExprPtr domains,IN CONST permutationPtr p,
				 IN COUNTER ilevel,IN COUNTER level,IN COUNTER shift));

DEF_PRIVATE_ACTION_1ARG(rendomk)
{
  struct renexprk *data;

  data = (struct renexprk *)ENV_UNDERLYING(env);
  return renameDomains(data->expr,data->p,data->ilevel,data->level,data->shift);
}
ACTION_END

PRIVATE ExprPtr renameDomainsk(domains,p,ilevel,level,shift)
IN CONST ExprPtr domains;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  struct renexprk *data = (struct renexprk *)alloctmp(sizeof(struct renexprk));

  data->expr = domains;
  data->p = p;
  data->ilevel = ilevel;
  data->level = level;
  data->shift = shift;
  return CallCC(ACTION_NAME(rendomk),(ExprPtr)data,true);
}

PRIVATE ExprPtr renameDomains(domains,p,ilevel,level,shift)
INOUT ExprPtr domains;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  ExprPtr aux,pat,e;

  if (domains!=NIL)
    {
      if (STKFULL(domains))
	return renameDomainsk(domains,p,ilevel,level,shift);
      switch (domains->kind)
	{
	case E_PAT_RECORD_GNRL_CONS:
	case E_PAT_REVRECORD_GNRL_CONS:
	case E_PAT_RECORD_MAPADD:
	  aux = CAR(domains);
	  if (shift!=0)
	    pat = renamePattern(CAR(aux),p,ilevel,level,shift,NIL);
	  else pat = CAR(aux); /* if shift==0, then pat cannot change. */
	  e = renameExpr(CDR(aux),p,ilevel,level,shift);
	  if (pat!=CAR(aux) || e!=CDR(aux))
	    {
	      aux = MakePair((ExprKind)aux->kind,pat,e);
	      (void) KEEP(aux);
	    }
	  e = renameDomains(CDR(domains),p,ilevel,level,shift);
	  if (aux!=CAR(domains) || e!=CDR(domains))
	    {
	      domains = MakePair((ExprKind)domains->kind,aux,e);
	      (void) KEEP(domains);
	    }
	  break;
	case E_TUPLE_1:
	  e = renameBindings(REFVAL(domains),p,ilevel,ilevel,shift);
	  /* Notice that the level given to renameBindings() is actually ilevel
	     (this is not a typo). */
	  if (e!=REFVAL(domains))
	    {
	      domains = MakeTuple1(e);
	      (void) KEEP(domains);
	    }
	  break;
	}
    }
  return domains;
}

PRIVATE ExprPtr renameTuple(e,p,ilevel,level,shift)
IN CONST ExprPtr e;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  COUNTER n,i;
  ExprPtr *elts;
  Boolean changed;

  changed = false;
  n = AVAL(e)->n;
  elts = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0; i<n; i++)
    {
      elts[i] = renameExpr(AVAL(e)->expr[i],p,ilevel,level,shift);
      if (!changed)
	changed = (elts[i]!=AVAL(e)->expr[i]);
    }
  if (changed)
    {
      e = MakeRecord((ExprKind)e->kind,ASHAPE(e),n,elts);
      (void) KEEP(e);
    }
  return e;
}

DEF_PRIVATE_ACTION_1ARG(renexprk)
{
  struct renexprk *data;

  data = (struct renexprk *)ENV_UNDERLYING(env);
  return renameExpr(data->expr,data->p,
		    data->ilevel,data->level,data->shift);
}
ACTION_END

PRIVATE ExprPtr renameExprk(expr,p,ilevel,level,shift)
IN CONST ExprPtr expr;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  struct renexprk *data = (struct renexprk *)alloctmp(sizeof(struct renexprk));

  data->expr = expr;
  data->p = p;
  data->ilevel = ilevel;
  data->level = level;
  data->shift = shift;
  return CallCC(ACTION_NAME(renexprk),(ExprPtr)data,true);
}

PRIVATE ExprPtr renameExpr(expr,p,ilevel,level,shift)
IN CONST ExprPtr expr;
IN CONST permutationPtr p;
IN COUNTER ilevel,level,shift;
{
  ExprPtr aux;
  ENTER;

  if (STKFULL(expr))
    return renameExprk(expr,p,ilevel,level,shift);
  /*if (expr==NIL)
    return false;*/
  switch (expr->kind)
    {
      case E_VARREF:
	return renameVarref(expr,p,level,shift);
      case E_REF: case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
      case E_GET: case E_RAISE: case E_DELAY:
      case E_NOT:
	aux = renameExpr(REFVAL(expr),p,ilevel,level,shift);
	if (aux!=REFVAL(expr))
	  {
	    expr = MakeCounted((ExprKind)expr->kind,aux,COUNTEDN(expr));
	  }
	break;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
	{
	  ExprPtr pat,e;

	  aux = COUNTEDVAL(expr);
	  e = renameExpr(CDR(aux),p,level+1,level+1,shift);
	  pat = renamePattern(CAR(aux),p,level+1,level,shift,NIL);
	  if (pat!=CAR(aux) || e!=CDR(aux))
	    {
	      aux = MakePair((ExprKind)aux->kind,pat,e);
	      (void) KEEP(aux);
	    }
	  if (aux!=COUNTEDVAL(expr))
	    {
	      expr = MakePolyfun(expr->magic,aux,COUNTEDN(expr));
	    }
	}
	break;
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
    case E_MEMO_FN:
      aux = renameExpr(CDR(expr),p,ilevel,level,shift);
      if (aux!=CDR(expr))
	{
	  expr = MakePair((ExprKind)expr->kind,CAR(expr),aux);
	}
      break;
    case E_PAT_AS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
    case E_MAPLET: case E_SET: case E_MAPGET:
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
    case E_IF: case E_PROG: case E_AND: case E_OR:
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE: case E_SET_WHILE:
    case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE:
    case E_INTER_WHILE: case E_MAP_WHILE: case E_MAP_UNDER_WHILE:
    case E_ALL_WHILE: case E_EXISTS_WHILE:
    case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
    case E_HANDLE: case E_OVERWRITE: case E_UNDERWRITE:
    case E_EQUAL: case E_NEQUAL:
      {
	ExprPtr car;

	car = renameExpr(CAR(expr),p,ilevel,level,shift);
	aux = renameExpr(CDR(expr),p,ilevel,level,shift);
	if (car!=CAR(expr) || aux!=CDR(expr))
	  {
	    expr = MakePair((ExprKind)expr->kind,car,aux);
	  }
      }
      break;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      expr = renameTuple(expr,p,ilevel,level,shift);
      break;
    case E_LET:
      {
	COUNTER newlevel = level+1;
	ExprPtr bind;

	if (IVAL(CAR(LETBIND(expr)))<0)
	  newlevel++;
	bind = renameBindings(CDR(LETBIND(expr)),p,newlevel,newlevel,shift);
	aux = renameExpr(LETBODY(expr),p,ilevel,newlevel,shift);
	if (bind!=CDR(LETBIND(expr)) || aux!=LETBODY(expr))
	  {
	    bind = MakePair(E_CONS,CAR(LETBIND(expr)),bind);
	    (void) KEEP(bind);
	    expr = MakeLet(bind,aux);
	  }
	break;
      }
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
    case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF: case E_EXISTS_OF:
    case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP:
    case E_INTER_MAP: case E_MAP_MAP: case E_MAP_UNDER_MAP:
    case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      {
	ExprPtr doms;

	ilevel = level+1;
	doms = renameDomains(CDR(OFDOM(expr)),p,ilevel,level,shift);
	aux = renameExpr(OFFP(expr),p,ilevel,ilevel,shift);
	/* level is set to ilevel, this is not a typo. */
	if (doms!=CDR(OFDOM(expr)) || aux!=OFFP(expr))
	  {
	    doms = MakePair(E_CONS,CAR(OFDOM(expr)),doms);
	    (void) KEEP(doms);
	    expr = MakePair((ExprKind)expr->kind,aux,doms);
	  }
	break;
      }
      case E_PACK:
    case E_DEBUGZONE:
      case E_PROFILE:
      aux = renameExpr(CAR(expr),p,ilevel,level,shift);
      if (aux!=CAR(expr))
	{
	  expr = MakePair((ExprKind)expr->kind,aux,CDR(expr));
	}
      break;
    case E_DATACON:
      break;
    }
  KEEPONLY(expr);
  return expr;
}

PUBLIC ExprPtr shiftExpr(e,shift)
IN CONST ExprPtr e;
IN COUNTER shift;
{
  if (shift!=0)
    e = renameExpr(e,nilPermutation(),0,0,shift);
  return e;
}

PRIVATE Boolean matchPattern P((IN CONST ExprPtr pat1,IN CONST ExprPtr pat2,
				INOUT permutationPtr p, INOUT INT *np));

#ifdef OBSOLETE
PRIVATE Boolean matchPatVector(pat1,pat2,p,np)
IN CONST ExprPtr pat1;
IN CONST ExprPtr pat2;
INOUT permutationPtr p;
INOUT INT *np;
{
  COUNTER i,n;
  INT saven;

  n = AVAL(pat1)->n;
  if (n!=AVAL(pat2)->n)
    return false;
  saven = *np;
  for (i=0; i<n; i++)
    if (!matchPattern(AVAL(pat1)->expr[i],
		      AVAL(pat2)->expr[i],p,np))
      {
	*np = saven;
	return false;
      }
  return true;
}
#endif

PRIVATE Boolean sameBits(b1,b2)
IN CONST REGISTER CHAR *b1,*b2;
{
  REGISTER COUNTER i;

  for (i=0; i<(E_UNUSED+7) >> 3; i++)
    if (*b1++ != *b2++)
      return false;
  return true;
}

PRIVATE MapPtr getRecLabels(pat,ellp)
IN CONST ExprPtr pat;
OUT Boolean *ellp;
{
  MapPtr map;
  ExprPtr xy;
  ENTER;

  map = NewMap();
  for (; /*pat!=NIL && */pat->kind==(CHAR)E_PAT_RECORD_MAPADD; pat=CDR(pat))
    {
      xy = CAR(pat);
      map = MapPut(CAR(xy),CDR(xy),map);
    }
  *ellp = (pat==patany);
  KEEPONLY(map);
  return map;
}

MAPSWEEPFUNC(do_matchpatrecmapadd)
{
  struct matchpatrecmapadd *data;

  data = MAPDATA(struct matchpatrecmapadd *);
  return !matchPattern(y,MapGetSure(x,data->map2),data->p,data->np);
}

PRIVATE Boolean matchPatRecMapAdd(pat1,pat2,p,np)
IN CONST ExprPtr pat1;
IN CONST ExprPtr pat2;
INOUT permutationPtr p;
INOUT INT *np;
{
  struct matchpatrecmapadd *data = (struct matchpatrecmapadd *)
    alloctmp(sizeof(struct matchpatrecmapadd));
  MapPtr map1;
  Boolean ell1,ell2,res;
  INT saven;

  map1 = getRecLabels(pat1,&ell1);
  data->map2 = getRecLabels(pat2,&ell2);
  if (ell1!=ell2 || !sameDomains(map1,data->map2))
    return false;
  data->p = p;
  saven = *np;
  res = !MapSweep(map1,do_matchpatrecmapadd,data);
  if (!res)
    *np = saven;
  return res;
}

DEF_PRIVATE_ACTION_1ARG(matchpatk)
{
  struct matchpatk *data;

  data = (struct matchpatk *)ENV_UNDERLYING(env);
  data->res = matchPattern(data->pat1,data->pat2,data->p,data->np);
  return NIL;
}
ACTION_END

PRIVATE Boolean matchPatternk(pat1,pat2,p,np)
IN CONST ExprPtr pat1;
IN CONST ExprPtr pat2;
INOUT permutationPtr p;
INOUT INT *np;
{
  struct matchpatk *data = (struct matchpatk *)alloctmp(sizeof(struct matchpatk));

  data->pat1 = pat1;
  data->pat2 = pat2;
  data->p = p;
  data->np = np;
  (void) CallCC(ACTION_NAME(matchpatk),(ExprPtr)data,true);
  return data->res;
}

PRIVATE Boolean matchPattern(pat1,pat2,p,np)
  /* tries to find a permutation of variables in pat2
     such that pat1 and pat2 recognize the same values,
     and every variable bound by pat2 is also
     bound by pat1.
     Patterns pat1 and pat2 are assumed to be of the same type.
  */
IN CONST ExprPtr pat1;
IN CONST ExprPtr pat2;
INOUT permutationPtr p;
INOUT INT *np;
{
  INT saven;

  if (STKFULL(pat1))
    return matchPatternk(pat1,pat2,p,np);
  saven = *np;
again:
  /*if (pat2==NIL) return pat1==NIL;*/
  switch (pat2->kind)
    {
    case E_VARINTRO:
      if (/*pat1==NIL || */pat1->kind!=(CHAR)E_VARINTRO)
	{
	fail:
	  *np = saven;
	  return false;
	}
      {
	COUNTER i;
	
	i = applyPermutation(p,VARRANK(pat2));
	extendPermutation(p,i,VARRANK(pat1));
	return true;
      }
    case E_DEBUGNAME:
      if (/*pat1==NIL || */pat1->kind!=(CHAR)E_DEBUGNAME)
	goto fail;
      pat1 = DNPAT(pat1);
      pat2 = DNPAT(pat2);
      goto again;
    case E_DATA:
      if (pat1->kind==pat2->kind &&
	  pat1->magic==pat2->magic && CAR(pat1)==CAR(pat2))
	{
	  pat1 = CDR(pat1);
	  pat2 = CDR(pat2);
	  goto again;
	}
      goto fail;
    case E_CONS: case E_UNSHARED_CONS: case E_PAT_REVCONS:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_CONS: case E_UNSHARED_CONS: case E_PAT_REVCONS:
	  if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	      matchPattern(CDR(pat1),CDR(pat2),p,np))
	    return true;
	}
      goto fail;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2: case E_PAT_REVTUPLE_2:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_TUPLE_2: case E_UNSHARED_TUPLE_2: case E_PAT_REVTUPLE_2:
	  if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	      matchPattern(CDR(pat1),CDR(pat2),p,np))
	    return true;
	}
      goto fail;
    case E_APPL:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_APPL:
	  if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	      matchPattern(CDR(pat1),CDR(pat2),p,np))
	    return true;
	}
      goto fail;
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
	  if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	      matchPattern(CDR(pat1),CDR(pat2),p,np))
	    return true;
	}
      goto fail;
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
	  if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	      matchPattern(CDR(pat1),CDR(pat2),p,np))
	    return true;
	}
      goto fail;
    case E_PAT_AS: case E_MAPLET:
    case E_PAT_MAPADD:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind)
	if (matchPattern(CAR(pat1),CAR(pat2),p,np) &&
	    matchPattern(CDR(pat1),CDR(pat2),p,np))
	  return true;
      /* simplistic, but efficient */
      goto fail;
    case E_PAT_RECORD_MAPADD:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind && matchPatRecMapAdd(pat1,pat2,p,np))
	return true;
      goto fail;
    case E_NIL:
    case E_TUPLE_0:
    case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_VARREF:
      if (pat1==pat2)
	return true;
      goto fail;
    case E_QUOTE:
      if (REFVAL(pat1)==REFVAL(pat2))
	return true;
      goto fail;
    case E_PAT_ANY:
patanyagain:
    /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_VARINTRO: case E_PAT_ANY:
	  {
	    COUNTER i;

	    i = applyPermutation(p,*np++);
	    extendPermutation(p,i,VARRANK(pat1));
	    return true;
	  }
	case E_DEBUGNAME:
	  pat1 = DNPAT(pat1);
	  goto patanyagain;
	}
      goto fail;
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
	  pat1 = REFVAL(pat1);
	  pat2 = REFVAL(pat2);
	  goto again;
	}
      goto fail;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_REF: case E_PAT_CONS:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind)
	{
	  pat1 = REFVAL(pat1);
	  pat2 = REFVAL(pat2);
	  goto again;
	}
      goto fail;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      /*if (pat1==NIL) return false;*/
      switch (pat1->kind)
	{
	case E_RECORD_1: case E_UNSHARED_RECORD_1:
	  if (CAR(pat1)!=CAR(pat2))
	    goto fail;
	  pat1 = CDR(pat1);
	  pat2 = CDR(pat2);
	  goto again;
	}
      goto fail;
    case E_PAT_CHECK:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind)
	{
	  if (!sameBits(PATCHECKBITS(pat1),PATCHECKBITS(pat2)))
	    goto fail;
	  pat1 = PATCHECKVAL(pat1);
	  pat2 = PATCHECKVAL(pat2);
	  goto again;
	}
      goto fail;
    case E_PAT_PACK:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind && PACKTYPE(pat1)==PACKTYPE(pat2))
	{
	  pat1 = PACKVAL(pat1);
	  pat2 = PACKVAL(pat2);
	  goto again;
	}
      goto fail;
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
      /*if (pat1==NIL) return false;*/
      if (pat1->kind==pat2->kind && COUNTEDN(pat1)==COUNTEDN(pat2))
	{
	  pat1 = COUNTEDVAL(pat1);
	  pat2 = COUNTEDVAL(pat2);
	  goto again;
	}
      goto fail;
      /*case E_PATCHOICE not mentioned */
    default:
      kinderror(message(633),pat2->kind);
      break;
    }
  return true;
}

PRIVATE ExprPtr reorderPatRecMapAdd(pats,pat,body,dbgnames,np)
IN CONST ExprPtr pats;
INOUT ExprPtr pat;
INOUT ExprPtr body;
INOUT ExprPtr dbgnames;
INOUT INT *np;
{ /* test whether the first maplet pattern in pats matches some
     maplet pattern in pat; if it is the case, reorder pat so
     that this pattern comes to the head of pat; otherwise,
     return pat. */
  permutationPtr p;
  ExprPtr lastpp,pp;

  for (lastpp = NIL, pp = pat;
	/*pp!=NIL && */pp->kind==(CHAR)E_PAT_RECORD_MAPADD;
	lastpp = pp, pp = CDR(pp))
    {
      p = nilPermutation();
      if (matchPattern(CAR(pats),CAR(pp),p,np))
	{ /* try to share the maplet pattern */
	  body = renameExpr(body,p,(COUNTER)0,(COUNTER)0,(COUNTER)0);
	  if (lastpp==NIL)
	    pat = CDR(pp);
	  else SETCDR(lastpp,CDR(pp)); /* erase maplet from list */
	  pat = renamePattern(pat,p,(COUNTER)0,(COUNTER)0,(COUNTER)0,dbgnames);
	  pat = tackPattern(CDR(pats),pat,body,dbgnames,np);
	  SETCDR(pp,pat);
	  return pp;
	}
    }
  {
    ExprPtr aux;

    MAKECONS(pat,body,aux);
    (void) KEEP(aux);
    pats = MakePair(E_PATCHOICE,pats,aux);
    (void) KEEP(pats);
  }
  return pats;
}

DEF_PRIVATE_ACTION_1ARG(tackpatk)
{
  struct tackpatk *data;

  data = (struct tackpatk *)ENV_UNDERLYING(env);
  return tackPattern(data->pats,data->pat,data->body,data->dbgnames,data->np);
}
ACTION_END

PRIVATE ExprPtr tackPatternk(pats,pat,body,dbgnames,np)
IN CONST ExprPtr pats;
IN CONST ExprPtr pat;
IN CONST ExprPtr body;
INOUT ExprPtr dbgnames;
INOUT INT *np;
{
  struct tackpatk *data = (struct tackpatk *)alloctmp(sizeof(struct tackpatk));

  data->pats = pats;
  data->pat = pat;
  data->body = body;
  data->dbgnames = dbgnames;
  data->np = np;
  return CallCC(ACTION_NAME(tackpatk),(ExprPtr)data,true);
}

MAPSWEEPFUNC(do_tackdbgnames)
{
  ExprPtr *patp;

  patp = MAPDATA(ExprPtr *);
  *patp = MakeDebugName(*patp,DNINFO(x));
  (void) KEEP(*patp);
  return false;
}

IMPORT ExprPtr MkQuote P((ExprPtr e));

PUBLIC ExprPtr MkQuote(e)
IN CONST ExprPtr e;
{
  switch (e->kind)
    {
    case E_NIL: case E_TUPLE_0: case E_RECORD_0: case E_EMPTY:
    case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      return e;
    default:
      return MakeQuote(e);
    }
}

PRIVATE ExprPtr unquoteData(val)
IN CONST ExprPtr val;
{
  if (val->magic<0)
    {
      val = MakeData(val->magic,CAR(val),NIL);
    }
  else
    {
      ExprPtr pat;

      pat = MakeQuote(CDR(val));
      (void) KEEP(pat);
      val = MakeData(val->magic,CAR(val),pat);
    }
  return KEEP(val);
}

PUBLIC ExprPtr tackPattern(pats,pat,body,dbgnames,np)
IN CONST ExprPtr pats;
IN CONST ExprPtr pat;
IN CONST ExprPtr body;
INOUT ExprPtr dbgnames;
INOUT INT *np;
{
  if (STKFULL(pats))
    return tackPatternk(pats,pat,body,dbgnames,np);
  /*if (pats==NIL) goto tackit;*/
  switch (pats->kind)
    {
    case E_PATCHOICE:
      {
	ExprPtr last,ppats;
	ExprPtr *patp;

	last = CDR(pats);
	patp = &CAR(last);
	while (ppats=*patp, ppats->kind==(CHAR)E_DEBUGNAME)
	  {
	    last = ppats;
	    patp = &DNPAT(last);
	  }
	pat = tackPattern(ppats,pat,body,dbgnames,np);
	SETQ(*patp,pat);
	HITBARRIER(last);
      }
      break;
    case E_CONS:
    case E_UNSHARED_CONS:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_CONS: case E_UNSHARED_CONS: case E_PAT_REVCONS:
	  {
	    permutationPtr p;
	    ExprPtr cdr;

tackcdr:
	    p = nilPermutation();
	    if (matchPattern(CAR(pats),CAR(pat),p,np))
	      { /* try to share the CAR pattern */
		body = renameExpr(body,p,(COUNTER)0,(COUNTER)0,(COUNTER)0);
		cdr = CDR(pat);
		cdr = renamePattern(cdr,p,(COUNTER)0,(COUNTER)0,(COUNTER)0,NIL);
		cdr = tackPattern(CDR(pats),cdr,body,dbgnames,np);
		pats = MakePair((ExprKind)pats->kind,CAR(pats),cdr);
		(void) KEEP(pats);
		break;
	      }
	    /*FALLTHROUGH*/
	  }
	default:
	  goto tackit;
	}
      break;
    case E_PAT_REVCONS:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_CONS: case E_UNSHARED_CONS: case E_PAT_REVCONS:
	  { /* try to share the CDR pattern */
	    permutationPtr p;
	    ExprPtr car;

tackcar:
	    p = nilPermutation();
	    if (matchPattern(CDR(pats),CDR(pat),p,np))
	      {
		body = renameExpr(body,p,(COUNTER)0,(COUNTER)0,(COUNTER)0);
		car = CAR(pat);
		car = renamePattern(car,p,(COUNTER)0,(COUNTER)0,(COUNTER)0,NIL);
		car = tackPattern(CAR(pats),car,body,dbgnames,np);
		pats = MakePair((ExprKind)pats->kind,car,CDR(pats));
		(void) KEEP(pats);
		break;
	      }
	    /*FALLTHROUGH*/
	  }
	default:
	  goto tackit;
	}
      break;
#ifdef DATACONPAT
    case E_DATACON: /*!!!*/
    datacon:
      switch (pat->kind)
	{
	case E_DATACON:
	  goto tackref;
	case E_QUOTE:
	  if (REFVAL(pat)->kind==(CHAR)E_DATA)
	    {
	      pat = unquoteData(REFVAL(pat));
	      pat = tackPattern(REFVAL(pats),pat,body,dbgnames,np);
	      SETREFVAL(pats,pat);
	    }
	  break;
	}
      goto tackit;
#endif
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_TUPLE_2: case E_UNSHARED_TUPLE_2: case E_PAT_REVTUPLE_2:
	  goto tackcdr;
	}
      goto tackit;
    case E_APPL:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_APPL:
	  goto tackcdr;
	}
      goto tackit;
    case E_PAT_REVTUPLE_2:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_TUPLE_2: case E_UNSHARED_TUPLE_2: case E_PAT_REVTUPLE_2:
	  goto tackcar;
	}
      goto tackit;
    case E_PAT_TUPLE_GNRL_CONS:
	/*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
	  goto tackcdr;
	  }
      goto tackit;
    case E_PAT_REVTUPLE_GNRL_CONS:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	  {
	  case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
	    goto tackcar;
	  }
      goto tackit;
    case E_PAT_RECORD_GNRL_CONS:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
	  goto tackcdr;
	}
      goto tackit;
    case E_PAT_REVRECORD_GNRL_CONS:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
	  goto tackcar;
	}
      goto tackit;
    case E_PAT_MAPADD:
      goto tackit; /* It is unsound to share PAT_MAPADD patterns: consider
		      {x => ()} U {SOME 2 => ()} and
		      {x => ()} U {NONE => ()};
		      only the second matches {NONE, SOME 3} (with x=SOME 3, y=NONE).
		      However if we share them, we shall get the
		      following pattern:
		      {x => ()} U PATCHOICE ({SOME 2 => ()}, {NONE => ()}),
		      for which the compiler (in bc.c) would first bind x to NONE,
		      then fail on both branches of the choice. (It does not know
		      how to backtrack to change its choice on x.)
		      Such constructs are however quite rare, so it is probably
		      not worth complicating the compiler (in bc.c) for just this. */
#ifdef OBSOLETE
      /*if (pat==NIL) goto tackit;*/
      if (pat->kind==(CHAR)E_PAT_MAPADD)
	{
	  goto tackcdr;
	}
      goto tackit;
#endif
    case E_PAT_RECORD_MAPADD:
      /*if (pat==NIL) goto tackit;*/
	if (pat->kind!=(CHAR)E_PAT_RECORD_MAPADD)
	  goto tackit;
	pats = reorderPatRecMapAdd(pats,pat,body,dbgnames,np);
      break;
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
	  goto tackref;
	}
      goto tackit;
    case E_PAT_CONS:
      /*if (pat==NIL) goto tackit;*/
      if (pat->kind==(CHAR)E_PAT_CONS)
	goto tackref;
      goto tackit;
    case E_REF:
      /*if (pat==NIL) goto tackit;*/
      if (pat->kind!=pats->kind)
	goto tackit;
tackref:
      pat = tackPattern(REFVAL(pats),REFVAL(pat),body,dbgnames,np);
      SETREFVAL(pats,pat);
      break;
    case E_QUOTE:
      if (REFVAL(pats)->kind!=(CHAR)E_DATA)
	goto tackit;
      pats = unquoteData(REFVAL(pats));
      /*FALLTHROUGH*/
    case E_DATA:
      switch (pat->kind)
	{
	case E_QUOTE:
	  if (REFVAL(pat)->kind!=(CHAR)E_DATA)
	    goto tackit;
	  pat = unquoteData(REFVAL(pat));
	  /*FALLTHROUGH*/
	case E_DATA:
	  if (pats->magic==pat->magic)
	    {
	    tack1cdr:
	      pat = tackPattern(CDR(pats),CDR(pat),body,dbgnames,np);
	      SETCDR(pats,pat);
	      break;
	    }
	  /*FALLTHROUGH*/
	default:
	  goto tackit;
	}
      break;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      /*if (pat==NIL) goto tackit;*/
      switch (pat->kind)
	{
	case E_RECORD_1: case E_UNSHARED_RECORD_1:
	  goto tack1cdr;
	}
      goto tackit;
    case E_PAT_CHECK:
	/*if (pat==NIL) goto tackit;*/
      if (pat->kind==(CHAR)E_PAT_CHECK && sameBits(PATCHECKBITS(pats),PATCHECKBITS(pat)))
	{
	  pat = tackPattern(PATCHECKVAL(pats),PATCHECKVAL(pat),body,dbgnames,np);
	  SETQ(PATCHECKVAL(pats),pat);
	  HITBARRIER(pats);
	  break;
	}
      goto tackit;
    case E_PAT_CARD_EQ:
      /*if (pat==NIL) goto tackit;*/
      if (pat->kind==(CHAR)E_PAT_CARD_EQ && COUNTEDN(pats)==COUNTEDN(pat))
	goto tackref;
      goto tackit;
    case E_PAT_CARD_GEQ:
      /*if (pat==NIL) goto tackit;*/
      if (pat->kind==(CHAR)E_PAT_CARD_GEQ && COUNTEDN(pats)==COUNTEDN(pat))
	goto tackref;
      goto tackit;
    default:
tackit:
      {
	ExprPtr aux;

	if (dbgnames!=NIL)
	  {
	    (void) MapSweep(REFVAL(dbgnames),do_tackdbgnames,&pat);
	    SETREFVAL(dbgnames,NewMap());
	  }
	MAKECONS(pat,body,aux);
	(void) KEEP(aux);
	pats = MakePair(E_PATCHOICE,pats,aux);
	(void) KEEP(pats);
	break;
      }
    }
  return pats;
}

DEF_PRIVATE_ACTION_1ARG(haschoice)
{
  struct haschoice *data;

  data = (struct haschoice *)ENV_UNDERLYING(env);
  data->res = hasChoice(data->pat);
  return NIL;
}
ACTION_END

PRIVATE Boolean hasChoice1(pat)
IN CONST ExprPtr pat;
{
  struct haschoice *d = (struct haschoice *)alloctmp(sizeof(struct haschoice));

  d->pat = pat;
  (void) CallCC(ACTION_NAME(haschoice),(ExprPtr)d,true);
  return d->res;
}

PUBLIC Boolean hasChoice(pat)
IN CONST ExprPtr pat;
{
  if (STKFULL(pat))
    return hasChoice1(pat);
again:
  switch (pat->kind)
    {
    case E_VARINTRO:
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_VARREF: case E_PAT_ANY:
      return false;
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_PATCHOICE:
      return true;
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_PAT_REVCONS: case E_PAT_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_APPL: case E_PAT_AS:
    case E_PAT_RECORD_MAPADD: case E_PAT_MAPADD:
      if (hasChoice(CAR(pat)))
	return true;
      pat = CDR(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1: case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_DATA:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    default:
      kinderror(message(665),pat->kind);
      /*NOTREACHED*/
      return false;
    }
}

DEF_PRIVATE_ACTION_1ARG(getdbgnames)
{
  struct getdbgnames *data;

  data = (struct getdbgnames *)ENV_UNDERLYING(env);
  data->res = getDbgNames(data->pat,data->res);
  return NIL;
}
ACTION_END

PRIVATE MapPtr getDbgNames1(pat,res)
     IN CONST ExprPtr pat;
     IN CONST MapPtr res;
{
  struct getdbgnames *d = (struct getdbgnames *)alloctmp(sizeof(struct getdbgnames));

  d->pat = pat;
  d->res = res;
  (void) CallCC(ACTION_NAME(getdbgnames),(ExprPtr)d,true);
  return d->res;
}

PUBLIC MapPtr getDbgNames(pat,res)
     IN CONST ExprPtr pat;
     IN CONST MapPtr res;
{
  ExprPtr info = NOEXPR;

  if (STKFULL(pat))
    return getDbgNames1(pat,res);
 again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      info = DNINFO(pat);
      pat = DNPAT(pat);
      goto again;
    case E_VARINTRO:
      if (info!=NOEXPR)
	{
	  res = MapPut (pat,info,res);
	  return res;
	}
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_VARREF: case E_PAT_ANY:
      return res;
    case E_PATCHOICE:
      /* not implemented: this case is undefined here.
	 That is, getDbgNames() should only be called on patterns
	 pat such that !hasChoice(pat). */
      return res;
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_PAT_REVCONS: case E_PAT_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_APPL: case E_PAT_AS:
    case E_PAT_RECORD_MAPADD: case E_PAT_MAPADD:
      res = getDbgNames (CAR(pat), res);
      pat = CDR(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1: case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_DATA:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    default:
      kinderror(message(948),pat->kind);
      /*NOTREACHED*/
      return res;
    }
}

