/* Miscellaneous support functions (memory allocation, floating-point arithmetic).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "portable.h"
#include "mem.h"
#include "memproto.h"
#ifdef THINK_C
#include <unix.h>
#endif
#ifdef HAS_PROTOTYPES
#include "smalloc.h"
#endif

#define __SEG__ EXPR

EXTERN Boolean initializing;
struct Expression;
IMPORT void GCgrow P((IN COUNTER urgency, IN CONST struct Expression *tag));

PUBLIC void safememcpy(to,from,size)
INOUT VOIDPTR to;
IN CONST VOIDPTR from;
IN INT size;
{
  if (size!=0) /* Some implementations of memcpy() don't like it when size==0... */
    memcpy(to,from,size);
}

#ifndef SBRK_MALLOC
#if SPECIAL_MALLOC
PRIVATE char * rainyDayFund=(char *)0L;
PRIVATE unsigned long rainyDaySize=0L;
#endif
#endif

PUBLIC void AllocRainyDayFund(size)
IN CONST unsigned long size;
{
#ifndef SBRK_MALLOC
#if SPECIAL_MALLOC
  while ((rainyDayFund = newptr_fail(size))==(char *)0L && size>100)
    size = (size*3)/4;
  rainyDaySize = MBrealsize(size);
#endif
#endif
}

#ifdef HAS_PROTOTYPES
PUBLIC char *ExpandFailMemHook(unsigned long size)
#else
PUBLIC char *ExpandFailMemHook(size)
unsigned long size;
#endif
{
  char * p;
#if SPECIAL_MALLOC
  unsigned long ssize;
  char * (*saveHook) P((unsigned long size));
  
  COMMITFREE();
  saveHook = theMZone->MZ_hook;
  theMZone->MZ_hook = MallocDefaultHook;
  p = Malloc(size);
  theMZone->MZ_hook = saveHook;
  if (p!=(char *)0L)
    return p;
#endif
  p = (char *)0L;
#if SPECIAL_MALLOC
  ssize = size;
  if (ssize<INC_SIZE)
    ssize = INC_SIZE;
#endif
#ifdef SBRK_MALLOC
  if (expandMZone(theMZone,ssize+MBSTRIDE)!=(MZPtr)0L)
    {
      p = ALLOC(size);
    }
#else
#if SPECIAL_MALLOC
  {
    unsigned long rsize,x;
    
    rsize = MBrealsize(ssize);
    if (rainyDaySize>=rsize && rainyDayFund!=(char *)0L)
      {
	if (rainyDaySize>=MBrealsize(0)+rsize)
	  {
	    x = rainyDaySize-rsize;
	    while (MBrealsize(x)+rsize>rainyDaySize)
	      x--;
	    (void) Mrealloc(rainyDayFund,x);
	    rainyDaySize = MBrealsize(x);
	  }
	else
	  {
	    rainyDaySize = 0;
	    FREE(rainyDayFund);
	    rainyDayFund = (char *)0L;
	  }
	p = ALLOC(size);
      }
  }
#endif
#endif
  return p;
}

#ifdef HAS_PROTOTYPES
PUBLIC char *ExpandMemHook(unsigned long size)
#else
PUBLIC char *ExpandMemHook(size)
unsigned long size;
#endif
{
  char * p;

  p = ExpandFailMemHook(size);
  if (p==(char *)0L)
    {
      nomem();
    }
  return p;
}

#ifdef HAS_PROTOTYPES
PUBLIC char *FailMemHook(unsigned long size)
#else
PUBLIC char *FailMemHook(size)
unsigned long size;
#endif
{
  char * p;

  p = ExpandFailMemHook(size);
  if (p==(char *)0L)
    {
      GCgrow(0, NULL);
      p = ALLOC(size);
    }
  return p;
}

#ifdef HAS_PROTOTYPES
PUBLIC char *GCMemHook(unsigned long size)
#else
PUBLIC char *GCMemHook(size)
unsigned long size;
#endif
{
  char * p;

  p = ExpandFailMemHook(size);
  if (p==(char *)0L)
    {
      GCgrow(0, NULL);
      p = newptr(size);
    }
  return p;
}

PUBLIC char *(*defaultMemHook) P((unsigned long size)) = GCMemHook;

#if !SPECIAL_MALLOC
#ifdef MALLOC_0_0
EXTERN char dummy;
#endif
#endif

PRIVATE char *hmalloc(size)
IN long size;
{
  REGISTER char *p;

#if !SPECIAL_MALLOC
  p = (char *)malloc(size);
  if (p==(char *)0)
    {
#ifdef MALLOC_0_0
      if (size==0) /* malloc(0) returns NULL on some ANSI implementations */
	return &dummy;
#endif
    }
#else
  p = Malloc(size);
#endif
  return p;
}

PUBLIC char *newptr(size)
long size;
{
  char *p;

#if !SPECIAL_MALLOC
  p = (char *)hmalloc(size);
  if (p==(char *)0)
    {
      p = ExpandMemHook(size);
    }
#else
  theMZone->MZ_hook = ExpandMemHook;
  p = hmalloc(size);
  theMZone->MZ_hook = defaultMemHook;
#endif
  return p;
}

PUBLIC char *newptr_gc(size)
long size;
{
  char *p;

#if !SPECIAL_MALLOC
  p = (char *)hmalloc(size);
  if (p==(char *)0)
    {
      p = GCMemHook(size);
    }
#else
  theMZone->MZ_hook = GCMemHook;
  p = hmalloc(size);
  theMZone->MZ_hook = defaultMemHook;
#endif
  return p;
}

PUBLIC char *new_gc(size)
long size;
{
  char *p;

  p = NEWPTR(size);
  if (p==NULL)
    {
#ifdef MALLOC_0_0
      if (size==0) /* malloc(0) returns NULL on some ANSI implementations */
	return &dummy;
#endif
      if (initializing)
	nomem();
      GCgrow(0, NULL);
      p = NEWPTR(size);
    }
  return p;
}

PUBLIC char *newptr_gc_fail(size)
long size;
{
  char *p;

#if !SPECIAL_MALLOC
  p = (char *)hmalloc(size);
  if (p==(char *)0)
    {
      p = FailMemHook(size);
    }
#else
  theMZone->MZ_hook = FailMemHook;
  p = hmalloc(size);
  theMZone->MZ_hook = defaultMemHook;
#endif
  return p;
}

PUBLIC char *newptr_fail(size)
long size;
{
  char *p;

#if !SPECIAL_MALLOC
  p = 
#ifdef MALLOC_0_0
    (size==0)?&dummy:
#endif
    (char *)hmalloc(size);
#else
  theMZone->MZ_hook = ExpandFailMemHook;
  p = hmalloc(size);
  theMZone->MZ_hook = defaultMemHook;
#endif
  return p;
}

PUBLIC char *malloc_fail(size)
long size;
{
#if !SPECIAL_MALLOC
  return
#ifdef MALLOC_0_0
    (size==0)?&dummy:
#endif
    (char *)hmalloc(size);
#else
  return hmalloc(size);
#endif
}

	/* try to get an unsigned type on 2 bytes in IEEE754 format */
#if FLOAT_FORMAT==IEEE754
#define xtype unsigned short
#define X ((xtype *)x)
#define Xx ((unsigned char *)x)

	/* note : quiet NaNs (C_QNAN) are ignored ! */

#if FLOAT_LAYOUT==1
PUBLIC int classifyD(x)
IN REGISTER CONST REAL *x;
{
  xtype x0;
  xtype neg;

  neg = (X[0] & 0x8000);
  x0 = X[0] & 0x7ff0;
  if (x0==0)
    {
      if (((X[0] & 0x7fff)==0) && X[1]==0 && X[2]==0 && X[3]==0)
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7ff0)
    {
      if (((X[0] & 0x7fff)==0x7ff0) && X[1]==0 && X[2]==0 && X[3]==0)
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==2
PUBLIC int classifyD(x)
IN REGISTER CONST REAL *x;
{
  xtype x0;
  xtype neg;

  neg = (X[0] & 0x0080);
  x0 = X[0] & 0xf07f;
  if (x0==0)
    {
      if (((X[0] & 0xff7f)==0) && X[1]==0 && X[2]==0 && X[3]==0)
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0xf07f)
    {
      if (((X[0] & 0xff7f)==0xf07f) && X[1]==0 && X[2]==0 && X[3]==0)
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==3
PUBLIC int classifyD(x)
IN REGISTER CONST REAL *x;
{
  unsigned char x0,x1;
  unsigned char neg;

  neg = (Xx[0] & 0x80);
  x0 = (Xx[0] & 0x7f);
  x1 = (Xx[1] & 0xf0);
  if (x0==0 && x1==0)
    {
      if (Xx[1]==0 && Xx[2]==0 && Xx[3]==0 && Xx[4]==0 && Xx[5]==0
	  && Xx[6]==0 && Xx[7]==0)
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7f && x1==0xf0)
    {
      if (Xx[1]==0xf0 && Xx[2]==0 && Xx[3]==0 && Xx[4]==0 && Xx[5]==0
	  && Xx[6]==0 && Xx[7]==0)
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==4
PUBLIC int classifyD(x)
IN REGISTER CONST REAL *x;
{
  xtype x0;
  xtype neg;

  neg = (X[3] & 0x8000);
  x0 = X[3] & 0x7ff0;
  if (x0==0)
    {
      if (((X[3] & 0x7fff)==0) && X[2]==0 && X[1]==0 && X[0]==0)
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7ff0)
    {
      if (((X[3] & 0x7fff)==0x7ff0) && X[2]==0 && X[1]==0 && X[0]==0)
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#endif
#endif
#endif
#endif

#if EXTP==1 || EXTP==2
#if FLOAT_LAYOUT==1
PUBLIC int classifyX(x)
IN REGISTER CONST EXTENDED *x;
{
  xtype x0;
  xtype neg;

  neg = X[0] & 0x8000;
  x0 = X[0] & 0x7fff;
  if (x0==0)
    {
      if (X[1]==0 && X[2]==0 && X[3]==0 && X[4]==0
#if EXTP==2
	  && X[5]==0
#endif
	  )
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7fff)
    {
      if (X[1]==0 && X[2]==0 && X[3]==0 && X[4]==0
#if EXTP==2
	  && X[5]==0
#endif
	  )
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==2
PUBLIC int classifyX(x)
IN REGISTER CONST EXTENDED *x;
{
  xtype x0;
  xtype neg;

  neg = X[0] & 0x0080;
  x0 = X[0] & 0xff7f;
  if (x0==0)
    {
      if (X[1]==0 && X[2]==0 && X[3]==0 && X[4]==0
#if EXTP==2
	  && X[5]==0
#endif
	  )
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0xff7f)
    {
      if (X[1]==0 && X[2]==0 && X[3]==0 && X[4]==0
#if EXTP==2
	  && X[5]==0
#endif
	  )
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==3
PUBLIC int classifyX(x)
IN REGISTER CONST EXTENDED *x;
{
  unsigned char x0,x1;
  unsigned char neg;

  neg = Xx[0] & 0x80;
  x0 = Xx[0] & 0x7f;
  x1 = Xx[1];
  if (x0==0 && x1==0)
    {
      if (Xx[2]==0 && Xx[3]==0 && Xx[4]==0 && Xx[5]==0 && Xx[6]==0
	  && Xx[7]==0 && Xx[8]==0 && Xx[9]==0
#if EXTP==2
	  && Xx[10]==0 && Xx[11]==0
#endif
	  )
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7f && x1=0xff)
    {
      if (Xx[2]==0 && Xx[3]==0 && Xx[4]==0 && Xx[5]==0 && Xx[6]==0
	  && Xx[7]==0 && Xx[8]==0 && Xx[9]==0
#if EXTP==2
	  && Xx[10]==0 && Xx[11]==0
#endif
	  )
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#else
#if FLOAT_LAYOUT==4
PUBLIC int classifyX(x)
IN REGISTER CONST EXTENDED *x;
{
  xtype x0;
  xtype neg;

  neg = X[3+EXTP] & 0x8000;
  x0 = X[3+EXTP] & 0x7fff;
  if (x0==0)
    {
      if (X[0]==0 && X[1]==0 && X[2]==0 && X[3]==0
#if EXTP==2
	  && X[4]==0
#endif
	  )
	return neg?(-C_ZERONUM):C_ZERONUM;
      return neg?(-C_DENORMALNUM):C_DENORMALNUM;
    }
  else if (x0==0x7fff)
    {
      if (X[0]==0 && X[1]==0 && X[2]==0 && X[3]==0
#if EXTP==2
	  && X[4]==0
#endif
	  )
	return neg?(-C_INFINITE):C_INFINITE;
      return neg?(-C_SNAN):C_SNAN;
    }
  return neg?(-C_NORMALNUM):C_NORMALNUM;
}
#endif
#endif
#endif
#endif
#endif

PUBLIC int NaNcodeD(x) /* *x must be a NaN */
IN REGISTER CONST REAL *x;
{
#if FLOAT_LAYOUT==1
  return (X[0] & 0x8000)?(-(int)(X[1]>>5)):(int)(X[1]>>5);
#else
#if FLOAT_LAYOUT==2
  return (X[0] & 0x80)?(-(int)((X[1]>>13) | ((X[1] & 0xff)<<3))):
    (int)((X[1]>>13) | ((X[1] & 0xff)<<3));
#else
#if FLOAT_LAYOUT==3
  return (Xx[0] & 0x80)?(-(int)(Xx[3]>>5)):(int)(Xx[3]>>5);
#else
#if FLOAT_LAYOUT==4
  return (X[3] & 0x8000)?(-(int)(X[2]>>5)):(int)(X[2]>>5);
#else
  return 0;
#endif
#endif
#endif
#endif
}

PUBLIC void makeNaND(i,x)
IN int i;
OUT REAL *x;
{
#if FLOAT_LAYOUT==1
  if (i<0)
    {
      X[0] = 0xfff8;
      i = -i;
    }
  else
    {
      X[0] = 0x7ff8;
    }
  X[1] = (i<<5);
  X[2] = X[3] = 0;
#else
#if FLOAT_LAYOUT==2
  if (i<0)
    {
      X[0] = 0xf8ff;
      i = -i;
    }
  else
    {
      X[0] = 0xf87f;
    }
  X[1] = (i<<13) | ((i>>3) & 0xff);
  X[2] = X[3] = 0;
#else
#if FLOAT_LAYOUT==3
  if (i<0)
    {
      Xx[0] = 0xff;
      Xx[1] = 0xf8;
      i = -i;
    }
  else
    {
      Xx[0] = 0x7f;
      Xx[1] = 0xf8;
    }
  Xx[3] = (i<<5);
  Xx[2] = Xx[4] = Xx[5] = Xx[6] = Xx[7] = 0;
#else
#if FLOAT_LAYOUT==4
  if (i<0)
    {
      X[3] = 0xfff8;
      i = -i;
    }
  else
    {
      X[3] = 0x7ff8;
    }
  X[2] = (i<<5);
  X[1] = X[0] = 0;
#else
  X[0] = X[1] = X[2] = X[3] = 0;
#endif
#endif
#endif
#endif
}

#else
PUBLIC int classifyD(x)
IN REGISTER CONST REAL *x;
{
  if ((*x)==0.0)
    return C_ZERONUM;
  if ((*x)==-0.0)
    return -C_ZERONUM;
  return ((*x)>0.0)?C_NORMALNUM:-C_NORMALNUM;
}

#if EXTP
PUBLIC int classifyX(x)
IN REGISTER CONST EXTENDED *x;
{
  if ((*x)==0.0)
    return C_ZERONUM;
  if ((*x)==-0.0)
    return -C_ZERONUM;
  return ((*x)>0.0)?C_NORMALNUM:-C_NORMALNUM;
}
#endif
#endif
