/* Page management for stop and copy GC.
   (Name refcount.c comes from the fact that there used to be reference counting functions here.)
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "special.h"
#include "cons.h"
#include "compile.h"
#include "type.h"
#include "errproto.h"

#ifdef HAS_PROTOTYPES
#include "srefcnt.h"
#endif

#define __SEG__	EXPR

#if STOP_AND_COPY_VECTORS
EXTERN Boolean initializing; /* during initialization, garbage collection
	is forbidden, so instead of allocating twin pages, we allocate
	single pages (we put them in twinPageList anyway) */
PUBLIC PagePtr twinPageList; /* list of couple of pages */
PUBLIC PagePtr curPage;
PUBLIC int spaceIndex = 0;
PUBLIC INT nPages = 0L,neededPages = 0L,allocedPages = 0L,minAllocedPages = 0L;

EXTERN SHORT generation_count;

EXTERN void fastmove P((long *from,long *to,unsigned long size));

EXTERN Boolean inGC;

PRIVATE PagePtr pageAlloc(next,new)
REGISTER PagePtr next; /* contains two pages (only one if initializing) */
IN allocFunction new;
{
  REGISTER PagePtr p;

  if (initializing)
    {
      p = (PagePtr)(*new)((unsigned long)sizeof(Page));
      if (p!=(PagePtr)0L)
	{
	  pageEnd(&p[0]) = (char *)&p[1]; /* correct because alignment
					     constraints are satisfied */
	  pageTop(&p[0]) = pageStart(&p[0]);
	  pageNext(&p[0]) = &next[0];
	  allocedPages++;
	}
    }
  else
    {
      p = (PagePtr)(*new)(2L*(unsigned long)sizeof(Page));
      if (p!=(PagePtr)0L)
	{
	  pageEnd(&p[0]) = (char *)&p[1]; /* correct because alignment
					     constraints are satisfied */
	  pageTop(&p[0]) = pageStart(&p[0]);
	  pageNext(&p[0]) = &next[0];
	  pageEnd(&p[1]) = (char *)&p[2]; /* correct because alignment
					     constraints are satisfied */
	  pageTop(&p[1]) = pageStart(&p[1]);
	  pageNext(&p[1]) = (next==(PagePtr)0L)?next:(&next[1]);
	  allocedPages++;
	}
    }
  return p;
}

PUBLIC PagePtr NewPage(new)
IN allocFunction new;
{
  PagePtr pp,pp0;

  if (curPage==(PagePtr)0L)
    {
      pp = pageAlloc(twinPageList,new);
      if (pp!=(PagePtr)0L)
	{
	  twinPageList = pp;
	  curPage = &pp[spaceIndex];
	}
    }
  else
    {
      pp0 = &curPage[-spaceIndex];
      pp = pageAlloc(pageNext(pp0),new);
      if (pp!=(PagePtr)0L)
	{
	  pageNext(&pp0[0]) = &pp[0];
	  if (!initializing)
	    pageNext(&pp0[1]) = &pp[1];
	}
    }
  return pp;
}

PUBLIC void DelPage()
{
  PagePtr pp0,pp1,pp2;

  pp0 = &curPage[-spaceIndex];
  pp1 = pageNext(pp0);
  pp2 = pageNext(pp1);
  MARKFREE(pp1);
  pageNext(&pp0[0]) = &pp2[0];
  if (!initializing)
    pageNext(&pp0[1]) = (pp2==(PagePtr)0L)?pp2:(&pp2[1]);
  allocedPages--;
}

PUBLIC char *nextscAlloc(rsize)
REGISTER unsigned long rsize;
{
  REGISTER char *p;
  PagePtr pp;
      /* rsize should be <=PAGESIZE, and actually <=MAXALLOC_IN_PAGE
	 to have less fragmentation across pages. */

  if (inGC)
    goto skipgc;
  if (rsize>MAXALLOC_IN_PAGE)
    return MALLOC(rsize);
  if ((curPage = pageNext(curPage))==(PagePtr)0L)
    {
      GCgrow(0, NULL);
      if (pageTop(curPage)+rsize<=pageEnd(curPage))
	{
	  p = pageTop(curPage);
	  pageTop(curPage) += rsize;
	  return p;
	}
skipgc:
      if ((pp = pageNext(curPage))==(PagePtr)0L)
	{
	  pp = NewPage(newptr_fail);
	  if (pp==(PagePtr)0L)
	    nomem();
	  curPage = pageNext(curPage);
	}
      else curPage = pp;
    }
  nPages++;
  p = pageStart(curPage);
  pageTop(curPage) = p+rsize;
  return p;
}

PUBLIC char *scCopy(p,size)
IN CONST REGISTER char *p;
IN REGISTER unsigned long size;
{
  REGISTER char *pp;

  if ((size = scRealSize(size))>MAXALLOC_IN_PAGE)
    return p; /* object was allocated by MALLOC(), must not be moved */
  scAlloc1(pp,size);
#if (ALIGNMENT % LONG_SIZE) == 0
  /*fastmove((long *)p,(long *)pp,size);*/
  {
    REGISTER long *to;

    to = (long *)pp;
    while (size!=0) /* works only if we never allocate 0-sized objects! */
      {
#if (ALIGNMENT % (2*LONG_SIZE))==0
#if (ALIGNMENT % (4*LONG_SIZE))==0
	size -= (4*LONG_SIZE);
	*to++ = ((long *)p)[0];
	*to++ = ((long *)p)[1];
	*to++ = ((long *)p)[2];
	*to++ = ((long *)p)[3];
	p += 4*LONG_SIZE;
#else
	size -= (2*LONG_SIZE);
	*to++ = ((long *)p)[0];
	*to++ = ((long *)p)[1];
	p += 2*LONG_SIZE;
#endif
#else
	size -= LONG_SIZE;
	*to++ = *(long *)p;
	p += LONG_SIZE;
#endif
      }
  }
#else
  safememcpy(pp,p,size);
#endif
  return pp;
}

#else
int dummy_variable_so_that_ar_does_not_complain_that_refcount_is_empty=0;
#endif
